from srb.core.action import ActionGroup, WheeledDriveActionCfg, WheeledDriveActionGroup
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, Transform, WheeledRobot
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class Perseverance(WheeledRobot):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/perseverance",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("rover")
            .joinpath("perseverance.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.02, rest_offset=0.005
            ),
            rigid_props=RigidBodyPropertiesCfg(
                max_linear_velocity=1.5,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=16,
                solver_velocity_iteration_count=4,
            ),
        ),
        actuators={
            "drive_joints": ImplicitActuatorCfg(
                joint_names_expr=["drive_joint.*"],
                velocity_limit=40.0,
                effort_limit=150.0,
                damping=25000.0,
                stiffness=0.0,
            ),
            "steer_joints": ImplicitActuatorCfg(
                joint_names_expr=["steer_joint.*"],
                velocity_limit=2.0,
                effort_limit=400.0,
                damping=200.0,
                stiffness=500.0,
            ),
            "rocker_joints": ImplicitActuatorCfg(
                joint_names_expr=["suspension_joint_rocker.*"],
                velocity_limit=5.0,
                effort_limit=2500.0,
                damping=400.0,
                stiffness=4000.0,
            ),
            "bogie_joints": ImplicitActuatorCfg(
                joint_names_expr=["suspension_joint_bogie.*"],
                velocity_limit=4.0,
                effort_limit=500.0,
                damping=25.0,
                stiffness=200.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = WheeledDriveActionGroup(
        WheeledDriveActionCfg(
            asset_name="robot",
            wheelbase=(2.26, 2.14764),
            wheelbase_mid=2.39164,
            wheel_radius=0.26268,
            steering_joint_names=[
                "steer_joint_front_left",
                "steer_joint_front_right",
                "steer_joint_rear_left",
                "steer_joint_rear_right",
            ],
            drive_joint_names=[
                "drive_joint_front_left",
                "drive_joint_front_right",
                "drive_joint_mid_left",
                "drive_joint_mid_right",
                "drive_joint_rear_left",
                "drive_joint_rear_right",
            ],
            scale_linear=0.7,
            scale_angular=deg_to_rad(75),
        )
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="body")
    frame_payload_mount: Frame = Frame(
        prim_relpath="body",
        offset=Transform(
            pos=(0.0, 0.0, 1.25),
            rot=rpy_to_quat(0.0, 0.0, -90.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="body",
        offset=Transform(
            pos=(0.0, -0.875, 1.25),
            rot=rpy_to_quat(0.0, 0.0, -90.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="body/camera_front",
        offset=Transform(
            # translation=(-0.3437, -0.8537, 1.9793),  # Left Navcam
            pos=(-0.7675, -0.8537, 1.9793),  # Right Navcam
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )



from .env import (  # noqa: F401
    OrbitalManipulationEnv,
    OrbitalManipulationEnvCfg,
    OrbitalManipulationEventCfg,
    OrbitalManipulationSceneCfg,
)
from .visual_ext import OrbitalManipulationEnvVisualExtCfg  # noqa: F401



from typing import TYPE_CHECKING, Tuple

import torch
from pydantic import BaseModel
from simforge import TexResConfig

from srb import assets
from srb.core.asset import AssetVariant, Object, RigidObjectCfg
from srb.core.domain import Domain
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.mdp import reset_root_state_uniform

if TYPE_CHECKING:
    from .task import TaskCfg


class SampleCfg(BaseModel, arbitrary_types_allowed=True):
    asset_cfg: RigidObjectCfg
    state_randomizer: EventTermCfg


def select_sample(
    env_cfg: "TaskCfg",
    *,
    prim_path: str = "{ENV_REGEX_NS}/sample",
    asset_cfg: SceneEntityCfg = SceneEntityCfg("sample"),
    seed: int = 0,
    init_state: RigidObjectCfg.InitialStateCfg = RigidObjectCfg.InitialStateCfg(),
    scale: Tuple[float, float, float] = (0.05, 0.05, 0.05),
    texture_resolution: TexResConfig | None = None,
    **kwargs,
) -> SampleCfg:
    pose_range = {
        "x": (-0.15, 0.15),
        "y": (-0.3, 0.3),
        "z": (0.1, 0.1),
        "roll": (-torch.pi, torch.pi),
        "pitch": (-torch.pi, torch.pi),
        "yaw": (-torch.pi, torch.pi),
    }

    if isinstance(env_cfg.sample, Object):
        if isinstance(env_cfg.sample, assets.SampleTube):
            pose_range.update(
                {
                    "z": (0.05, 0.05),
                    "roll": (torch.pi / 7, torch.pi / 7),
                    "pitch": (
                        87.5 * torch.pi / 180,
                        87.5 * torch.pi / 180,
                    ),
                    "yaw": (-torch.pi, torch.pi),
                }
            )
        sample_cfg = env_cfg.sample.asset_cfg.copy()  # type: ignore
        assert isinstance(sample_cfg, RigidObjectCfg)

    match env_cfg.sample:
        case AssetVariant.PRIMITIVE:
            pose_range["z"] = (0.1, 0.1)
            sample_cfg = assets.RandomShape(scale=scale).asset_cfg

        case AssetVariant.DATASET:
            match env_cfg.domain:
                case Domain.MARS:
                    pose_range.update(
                        {
                            "z": (0.05, 0.05),
                            "roll": (torch.pi / 7, torch.pi / 7),
                            "pitch": (
                                87.5 * torch.pi / 180,
                                87.5 * torch.pi / 180,
                            ),
                            "yaw": (-torch.pi, torch.pi),
                        }
                    )
                    sample_cfg = assets.SampleTube().asset_cfg
                case _:
                    pose_range["z"] = (0.07, 0.07)
                    sample_cfg = assets.ShortProfilePeg().asset_cfg

        case AssetVariant.PROCEDURAL:
            pose_range["z"] = (0.06, 0.06)
            match env_cfg.domain:
                case Domain.MOON:
                    sample_cfg = assets.LunarRock(
                        scale=scale, texture_resolution=texture_resolution
                    ).asset_cfg

                case Domain.MARS:
                    sample_cfg = assets.MarsRock(
                        scale=scale, texture_resolution=texture_resolution
                    ).asset_cfg

            sample_cfg.spawn.seed = seed  # type: ignore

    sample_cfg.prim_path = prim_path
    sample_cfg.init_state = init_state
    sample_cfg.spawn.replace(**kwargs)  # type: ignore

    return SampleCfg(
        asset_cfg=sample_cfg,
        state_randomizer=EventTermCfg(
            func=reset_root_state_uniform,
            mode="reset",
            params={
                "asset_cfg": asset_cfg,
                "pose_range": pose_range,
                "velocity_range": {},
            },
        ),
    )



import contextlib
import os
import sys
from functools import wraps
from typing import Any, Callable, TextIO


class StreamRedirect(contextlib.ContextDecorator):
    def __init__(self, source: TextIO | Any, target: str | TextIO | Any):
        self._target = target
        self._source_fd = source.fileno()
        self._dup_source_fd = None
        self._stream = None

    def __enter__(self):
        self._dup_source_fd = os.dup(self._source_fd)
        if isinstance(self._target, str):
            self._stream = open(self._target, "w")
        else:
            self._stream = self._target
        os.dup2(self._stream.fileno(), self._source_fd)

    def __exit__(self, exc_type, exc_value, traceback):
        os.dup2(
            self._dup_source_fd,  # type: ignore
            self._source_fd,
        )
        os.close(
            self._dup_source_fd,  # type: ignore
        )
        if isinstance(self._target, str) and self._stream is not None:
            self._stream.close()


class StdoutSuppress(StreamRedirect):
    def __init__(self):
        super().__init__(source=sys.stdout, target=os.devnull)


class StderrSuppress(StreamRedirect):
    def __init__(self):
        super().__init__(source=sys.stderr, target=os.devnull)


def suppress_stdout(fn: Callable):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        with StdoutSuppress():
            return fn(*args, **kwargs)

    return wrapper


def suppress_stderr(fn: Callable):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        with StderrSuppress():
            return fn(*args, **kwargs)

    return wrapper



from enum import Enum, auto

from typing_extensions import Self


class ObjectType(str, Enum):
    COMMON = auto()
    LIGHT = auto()
    PAYLOAD = auto()
    PEDESTAL = auto()
    TOOL = auto()

    def __str__(self) -> str:
        return self.name.lower()

    @classmethod
    def from_str(cls, string: str) -> Self | None:
        return next(
            (variant for variant in cls if string.upper() == variant.name), None
        )



from .base import (  # noqa: F401
    BaseEnvCfg,
    BaseEventCfg,
    BaseSceneCfg,
    DirectEnv,
    DirectEnvCfg,
    DirectMarlEnv,
    DirectMarlEnvCfg,
    ManagedEnv,
    ManagedEnvCfg,
)
from .extension import VisualExt, VisualExtCfg  # noqa: F401



from typing import TYPE_CHECKING, Tuple

from simforge import TexResConfig

from srb import assets
from srb.core.asset import RigidObjectCfg

if TYPE_CHECKING:
    from .task import TaskCfg


def select_obstacle(
    env_cfg: "TaskCfg",
    *,
    prim_path: str = "{ENV_REGEX_NS}/obstacle",
    seed: int = 0,
    init_state: RigidObjectCfg.InitialStateCfg = RigidObjectCfg.InitialStateCfg(),
    scale: Tuple[float, float, float] = (5.0, 5.0, 5.0),
    texture_resolution: TexResConfig | None = None,
    **kwargs,
) -> RigidObjectCfg:
    obstacle_cfg = assets.Asteroid(
        scale=scale, texture_resolution=texture_resolution
    ).asset_cfg

    obstacle_cfg.spawn.seed = seed  # type: ignore
    obstacle_cfg.prim_path = prim_path
    obstacle_cfg.init_state = init_state
    obstacle_cfg.spawn.replace(**kwargs)  # type: ignore

    obstacle_cfg.spawn.assets[0].geo.ops[0].scale_std = (  # type: ignore
        0.2 * scale[0],
        0.2 * scale[1],
        0.2 * scale[2],
    )

    return obstacle_cfg



from typing import Tuple

import numpy
from isaacsim.core.prims import SingleXFormPrim
from isaacsim.core.utils.articulations import move_articulation_root
from isaacsim.core.utils.prims import (
    get_articulation_root_api_prim_path,
    get_prim_at_path,
    is_prim_path_valid,
)
from isaacsim.core.utils.string import find_unique_string_name
from isaacsim.robot_setup.assembler import RobotAssembler as __RobotAssembler
from pxr import Usd, UsdPhysics
from pydantic import BaseModel

from .assembled_bodies import AssembledBodies
from .assembled_robot import AssembledRobot


class RobotAssemblerCfg(BaseModel):
    base_path: str
    attach_path: str
    base_mount_frame: str = ""
    attach_mount_frame: str = ""
    fixed_joint_offset: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    fixed_joint_orient: Tuple[float, float, float, float] = (1.0, 0.0, 0.0, 0.0)
    mask_all_collisions: bool = False
    mask_attached_collisions: bool = True
    disable_root_joints: bool = True


class RobotAssembler(__RobotAssembler):
    def assemble_rigid_bodies(self, cfg: RobotAssemblerCfg) -> AssembledBodies:
        fixed_joint_offset = numpy.array(cfg.fixed_joint_offset)
        fixed_joint_orient = numpy.array(cfg.fixed_joint_orient)

        # Make mount_frames if they are not specified
        if cfg.base_mount_frame:
            base_mount_path = (
                f"{cfg.base_path}/{cfg.base_mount_frame.removeprefix('/')}"
            )
        else:
            base_mount_path = (
                f"{cfg.base_path}/{cfg.attach_path.split('/')[-1]}_mount_frame"
            )
            base_mount_path = find_unique_string_name(
                base_mount_path, lambda x: not is_prim_path_valid(x)
            )
            SingleXFormPrim(base_mount_path, translation=(0, 0, 0))

        if cfg.attach_mount_frame:
            attach_mount_path = (
                f"{cfg.attach_path}/{cfg.attach_mount_frame.removeprefix('/')}"
            )
        else:
            attach_mount_path = (
                f"{cfg.attach_path}/{cfg.base_path.split('/')[-1]}_mount_frame"
            )
            attach_mount_path = find_unique_string_name(
                attach_mount_path, lambda x: not is_prim_path_valid(x)
            )
            SingleXFormPrim(attach_mount_path, translation=(0, 0, 0))

        # Get the prim and articulation root of the attached asset
        attach_prim = get_prim_at_path(cfg.attach_path)
        articulation_root = get_prim_at_path(
            get_articulation_root_api_prim_path(cfg.attach_path)
        )

        # Move the Articulation root to the attach path to avoid edge cases with physics parsing.
        if articulation_root.HasAPI(UsdPhysics.ArticulationRootAPI):  # type: ignore
            move_articulation_root(articulation_root, attach_prim)

        # Find and Disable Fixed Joints that Tie Object B to the Stage
        root_joints = [p for p in Usd.PrimRange(attach_prim) if self.is_root_joint(p)]

        if cfg.disable_root_joints:
            for root_joint in root_joints:
                root_joint.GetProperty("physics:jointEnabled").Set(False)

        if attach_prim.HasAttribute("physics:kinematicEnabled"):
            attach_prim.GetAttribute("physics:kinematicEnabled").Set(False)  # type: ignore

        # Create fixed Joint between attach frames
        fixed_joint = self.create_fixed_joint(
            attach_mount_path,
            base_mount_path,
            attach_mount_path,
            fixed_joint_offset,
            fixed_joint_orient,
        )

        # Make sure that Articulation B is not parsed as a part of Articulation A.
        fixed_joint.GetExcludeFromArticulationAttr().Set(True)

        # Mask collisions
        if cfg.mask_all_collisions:
            # base_path_art_root = get_articulation_root_api_prim_path(cfg.base_path)
            collision_mask = self.mask_collisions(cfg.base_path, cfg.attach_path)
        elif cfg.mask_attached_collisions:
            collision_mask = self.mask_collisions(base_mount_path, attach_mount_path)
        else:
            collision_mask = None

        return AssembledBodies(
            cfg.base_path,
            cfg.attach_path,
            fixed_joint,
            root_joints,
            articulation_root,
            collision_mask,
        )

    def assemble_articulations(
        self, cfg: RobotAssemblerCfg, single_robot: bool = False
    ) -> AssembledRobot:
        assemblage = self.assemble_rigid_bodies(cfg=cfg)

        if single_robot:
            art_b_prim = get_prim_at_path(cfg.attach_path)
            if art_b_prim.HasProperty("physxArticulation:articulationEnabled"):
                art_b_prim.GetProperty("physxArticulation:articulationEnabled").Set(
                    False
                )
            assemblage.fixed_joint.GetExcludeFromArticulationAttr().Set(False)

        return AssembledRobot(assemblage)



from srb.core.asset import AssetBaseCfg, ExtravehicularScenery
from srb.core.sim import CollisionPropertiesCfg, UsdFileCfg
from srb.utils.math import rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class StaticGateway(ExtravehicularScenery):
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/gateway",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("spacecraft")
            .joinpath("gateway.usdz")
            .as_posix(),
            collision_props=CollisionPropertiesCfg(),
        ),
    )


class StaticVenusExpress(ExtravehicularScenery):
    ## Model
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/venus_express",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("spacecraft")
            .joinpath("venus_express.usdz")
            .as_posix(),
            collision_props=CollisionPropertiesCfg(),
        ),
        init_state=AssetBaseCfg.InitialStateCfg(
            pos=(-0.55, 0.0, -0.35), rot=rpy_to_quat(0.0, 0.0, 90.0)
        ),
    )



from srb.core.action import ActionGroup, JointPositionActionGroup
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ActiveTool, ArticulationCfg, Frame, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    FixedTendonPropertiesCfg,
    JointDrivePropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.nucleus import ISAAC_NUCLEUS_DIR


class ShadowHand(ActiveTool):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/shadow_hand",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/ShadowHand/shadow_hand_instanceable.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                retain_accelerations=True,
                max_depenetration_velocity=1000.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=0,
            ),
            joint_drive_props=JointDrivePropertiesCfg(drive_type="force"),
            fixed_tendons_props=FixedTendonPropertiesCfg(
                limit_stiffness=30.0, damping=0.1
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.5),
            rot=(0.0, 0.0, -0.7071, 0.7071),
            joint_pos={".*": 0.0},
        ),
        actuators={
            "fingers": ImplicitActuatorCfg(
                joint_names_expr=[
                    "robot0_WR.*",
                    "robot0_(FF|MF|RF|LF|TH)J(3|2|1)",
                    "robot0_(LF|TH)J4",
                    "robot0_THJ0",
                ],
                effort_limit={
                    "robot0_WRJ1": 4.785,
                    "robot0_WRJ0": 2.175,
                    "robot0_(FF|MF|RF|LF)J1": 0.7245,
                    "robot0_FFJ(3|2)": 0.9,
                    "robot0_MFJ(3|2)": 0.9,
                    "robot0_RFJ(3|2)": 0.9,
                    "robot0_LFJ(4|3|2)": 0.9,
                    "robot0_THJ4": 2.3722,
                    "robot0_THJ3": 1.45,
                    "robot0_THJ(2|1)": 0.99,
                    "robot0_THJ0": 0.81,
                },
                stiffness={
                    "robot0_WRJ.*": 5.0,
                    "robot0_(FF|MF|RF|LF|TH)J(3|2|1)": 1.0,
                    "robot0_(LF|TH)J4": 1.0,
                    "robot0_THJ0": 1.0,
                },
                damping={
                    "robot0_WRJ.*": 0.5,
                    "robot0_(FF|MF|RF|LF|TH)J(3|2|1)": 0.1,
                    "robot0_(LF|TH)J4": 0.1,
                    "robot0_THJ0": 0.1,
                },
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup()

    ## Frames
    frame_mount: Frame = Frame(
        prim_relpath="robot0_forearm", offset=Transform(pos=(0.0, 0.0, -0.04))
    )
    frame_tool_centre_point: Frame = Frame(
        prim_relpath="robot0_forearm", offset=Transform(pos=(0.0, -0.025, 0.3))
    )



import functools
from typing import TYPE_CHECKING, Any, Dict, List, Mapping

import embodied
import gymnasium
import numpy
import torch
from elements import Space

if TYPE_CHECKING:
    from srb._typing import AnyEnv


class EmbodiedEnvWrapper(embodied.Env):
    def __init__(
        self,
        env: "AnyEnv",
        obs_key: str = "image",
        act_key: str = "action",
    ):
        self.env = env
        self._obs_key = obs_key
        self._act_key = act_key

        # Extract spaces
        self._obs_space = self.unwrapped.single_observation_space  # type: ignore
        self._action_space = self.unwrapped.single_action_space  # type: ignore
        self._is_obs_dict = hasattr(self._obs_space, "spaces")
        self._is_act_dict = hasattr(self._action_space, "spaces")

        # Extract useful information
        self._num_envs = self.unwrapped.num_envs
        self._device = self.unwrapped.device

        # Init buffers
        self._done = numpy.ones(self._num_envs, dtype=bool)
        self._info = [None for _ in range(self._num_envs)]
        self._ep_len_buf = torch.zeros(self._num_envs, device=self._device)
        self._ep_rew_buf = torch.zeros(self._num_envs, device=self._device)

    def __len__(self) -> int:
        return self._num_envs

    def __str__(self) -> str:
        return f"<{type(self).__name__}{self.env}>"

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def class_name(cls) -> str:
        return cls.__name__

    @property
    def unwrapped(self) -> "AnyEnv":
        return self.env.unwrapped  # type: ignore

    def get_episode_rewards(self) -> List[float]:
        return self._ep_rew_buf.cpu().tolist()

    def get_episode_lengths(self) -> List[int]:
        return self._ep_len_buf.cpu().tolist()

    @property
    def info(self) -> Mapping[str, Any]:
        return self._info  # type: ignore

    @functools.cached_property
    def obs_space(self):
        if self._is_obs_dict:
            spaces = self._flatten(self.unwrapped.single_observation_space.spaces)  # type: ignore
        else:
            spaces = {self._obs_key: self.unwrapped.single_observation_space}  # type: ignore
        spaces = {k: self._convert(v) for k, v in spaces.items()}
        return {
            **spaces,
            "reward": Space(numpy.float32),
            "is_first": Space(bool),
            "is_last": Space(bool),
            "is_terminal": Space(bool),
        }

    @functools.cached_property
    def act_space(self):
        if self._is_act_dict:
            spaces = self._flatten(self._action_space.spaces)  # type: ignore
        else:
            spaces = {self._act_key: self._action_space}
        spaces = {k: self._convert(v) for k, v in spaces.items()}
        spaces["reset"] = Space(bool)
        return spaces

    def seed(self, seed: int | None = None) -> List[int | None]:
        return [self.unwrapped.seed(seed)] * self.unwrapped.num_envs  # type: ignore

    def reset(
        self,
    ) -> Mapping[str, numpy.ndarray]:
        obs, self._info = self.env.reset()
        self._done = numpy.zeros(self._num_envs, dtype=bool)
        return self._obs(
            obs=obs,  # type: ignore
            reward=numpy.zeros(self._num_envs, dtype=numpy.float32),
            is_first=numpy.ones(self._num_envs, dtype=bool),
            is_last=numpy.zeros(self._num_envs, dtype=bool),
            is_terminal=numpy.zeros(self._num_envs, dtype=bool),
        )

    def step(self, action: Mapping[str, torch.Tensor]) -> Mapping[str, numpy.ndarray]:
        if action["reset"].all() or self._done.all():
            return self.reset()

        if self._is_act_dict:
            act = self._unflatten(action)  # type: ignore
        else:
            act = action[self._act_key]  # type: ignore
            if not isinstance(act, torch.Tensor):
                act = torch.from_numpy(act).to(device=self._device)
            else:
                act = act.to(device=self._device)

        obs, reward, terminated, truncated, self._info = self.env.step(act)  # type: ignore

        self._ep_rew_buf += reward
        self._ep_len_buf += 1

        self._done = (terminated | truncated).detach().cpu()  # type: ignore
        reset_ids = (self._done > 0).nonzero()

        reward = reward.detach().cpu()  # type: ignore
        terminated = terminated.detach().cpu()  # type: ignore
        truncated = truncated.detach().cpu()  # type: ignore

        self._ep_rew_buf[reset_ids] = 0
        self._ep_len_buf[reset_ids] = 0

        return self._obs(
            obs=obs,  # type: ignore
            reward=reward,
            is_first=numpy.zeros(self._num_envs, dtype=bool),
            is_last=self._done,
            is_terminal=terminated,
        )

    def render(self) -> numpy.ndarray:
        image = self.env.render()
        assert image is not None
        return image

    def close(self):
        try:
            self.env.close()
        except Exception:
            pass

    def get_attr(self, attr_name, indices=None) -> Any:
        if indices is None:
            indices = slice(None)
            num_indices = self._num_envs
        else:
            num_indices = len(indices)
        attr_val = getattr(self.env, attr_name)
        if not isinstance(attr_val, torch.Tensor):
            return [attr_val] * num_indices
        else:
            return attr_val[indices].detach().cpu()

    def set_attr(self, attr_name, value, indices=None):
        raise NotImplementedError("Setting attributes is not supported.")

    def env_method(
        self, method_name: str, *method_args, indices=None, **method_kwargs
    ) -> Any:
        if method_name == "render":
            return self.env.render()
        else:
            env_method = getattr(self.env, method_name)
            return env_method(*method_args, indices=indices, **method_kwargs)

    def _obs(
        self,
        obs: numpy.ndarray | torch.Tensor | dict[str, numpy.ndarray | torch.Tensor],
        reward: numpy.ndarray | torch.Tensor,
        is_first: numpy.ndarray | torch.Tensor,
        is_last: numpy.ndarray | torch.Tensor,
        is_terminal: numpy.ndarray | torch.Tensor,
    ) -> Mapping[str, numpy.ndarray | torch.Tensor]:
        if not self._is_obs_dict:
            obs = {self._obs_key: obs}  # type: ignore
        obs = self._flatten(obs)
        obs = {k: v.detach().cpu() for k, v in obs.items()}  # type: ignore
        obs.update(
            reward=reward,
            is_first=is_first,
            is_last=is_last,
            is_terminal=is_terminal,
        )
        return obs

    def _flatten(self, nest, prefix=None) -> Dict[str, Any]:
        result = {}
        for key, value in nest.items():
            key = prefix + "/" + key if prefix else key
            if isinstance(value, gymnasium.spaces.Dict):
                value = value.spaces
            if isinstance(value, Dict):
                result.update(self._flatten(value, key))
            else:
                result[key] = value
        return result

    def _unflatten(self, flat) -> Dict[str, Any]:
        result = {}
        for key, value in flat.items():
            parts = key.split("/")
            node = result
            for part in parts[:-1]:
                if part not in node:
                    node[part] = {}
                node = node[part]
            node[parts[-1]] = value
        return result

    def _convert(self, space) -> Space:
        if hasattr(space, "n"):
            return Space(numpy.int32, (), 0, space.n)
        return Space(space.dtype, space.shape, space.low, space.high)



from collections import defaultdict

import elements
import numpy
from isaacsim.simulation_app import SimulationApp

from srb.integrations.dreamer.driver import DriverParallelEnv


def eval_only(
    make_agent,
    make_env,
    make_logger,
    args,
    sim_app: SimulationApp,
):
    assert args.from_checkpoint

    agent = make_agent()
    logger = make_logger()

    logdir = elements.Path(args.logdir)
    logdir.mkdir()
    print("Logdir", logdir)
    step = logger.step
    usage = elements.Usage(**args.usage)
    agg = elements.Agg()
    epstats = elements.Agg()
    episodes = defaultdict(elements.Agg)
    should_log = elements.when.Clock(args.log_every)
    policy_fps = elements.FPS()

    @elements.timer.section("logfn")
    def logfn(tran, worker):
        episode = episodes[worker]
        tran["is_first"] and episode.reset()  # type: ignore
        episode.add("score", tran["reward"], agg="sum")
        episode.add("length", 1, agg="sum")
        episode.add("rewards", tran["reward"], agg="stack")
        for key, value in tran.items():
            isimage = (value.dtype == numpy.uint8) and (value.ndim == 3)
            if isimage and worker == 0:
                episode.add(f"policy_{key}", value, agg="stack")
            elif key.startswith("log/"):
                assert value.ndim == 0, (key, value.shape, value.dtype)
                episode.add(key, value, agg=("avg", "max", "sum"))  # type: ignore
        if tran["is_last"]:
            result = episode.result()
            logger.add(
                {
                    "score": result.pop("score"),
                    "length": result.pop("length"),
                },
                prefix="episode",
            )
            rew = result.pop("rewards")
            if len(rew) > 1:
                result["reward_rate"] = (numpy.abs(rew[1:] - rew[:-1]) >= 0.01).mean()
            epstats.add(result)

    ########### ONLY THIS IS CHANGED ############
    # fns = [bind(make_env, i) for i in range(args.envs)]
    # driver = embodied.Driver(fns, parallel=(not args.debug))
    driver = DriverParallelEnv(env=make_env(), num_envs=args.envs)
    ########### ONLY THIS IS CHANGED ############
    driver.on_step(lambda tran, _: step.increment())
    driver.on_step(lambda tran, _: policy_fps.step())
    driver.on_step(logfn)

    cp = elements.Checkpoint()
    cp.agent = agent
    cp.load(args.from_checkpoint, keys=["agent"])

    print("Start evaluation")
    policy = lambda *args: agent.policy(*args, mode="eval")  # noqa: E731
    driver.reset(agent.init_policy)
    while step < args.steps:
        ########### Sim extra (optional) ############
        if not sim_app.is_running():
            break
        ########### Sim extra (optional) ############
        driver(policy, steps=10)
        if should_log(step):
            logger.add(agg.result())
            logger.add(epstats.result(), prefix="epstats")
            logger.add(usage.stats(), prefix="usage")
            logger.add({"fps/policy": policy_fps.result()})
            logger.add({"timer": elements.timer.stats()["summary"]})
            logger.write()

    logger.close()



from typing import Sequence

import torch

from srb._typing import StepReturn
from srb.core.env import GroundEnv, GroundEnvCfg, GroundEventCfg, GroundSceneCfg
from srb.core.manager import EventTermCfg
from srb.core.marker import ARROW_CFG, VisualizationMarkers
from srb.core.mdp import randomize_command
from srb.core.sim import PreviewSurfaceCfg
from srb.utils.cfg import configclass
from srb.utils.math import matrix_from_quat, rotmat_to_rot6d

##############
### Config ###
##############


@configclass
class SceneCfg(GroundSceneCfg):
    pass


@configclass
class EventCfg(GroundEventCfg):
    command = EventTermCfg(
        func=randomize_command,
        mode="interval",
        interval_range_s=(0.5, 5.0),
        params={
            "env_attr_name": "_command",
            # "magnitude": 1.0,
        },
    )


@configclass
class TaskCfg(GroundEnvCfg):
    ## Scene
    scene: SceneCfg = SceneCfg()
    stack: bool = True

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    episode_length_s: float = 20.0
    is_finite_horizon: bool = False

    ## Visualization
    command_vis: bool = True

    def __post_init__(self):
        super().__post_init__()


############
### Task ###
############


class Task(GroundEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Initialize buffers
        self._command = torch.zeros(self.num_envs, 3, device=self.device)

        ## Visualization
        if self.cfg.command_vis:
            self._setup_visualization_markers()

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)

    def _setup_visualization_markers(self):
        ## Linear velocity
        cfg = ARROW_CFG.copy().replace(  # type: ignore
            prim_path="/Visuals/command/target_linvel"
        )
        cfg.markers["arrow"].tail_radius = 0.01
        cfg.markers["arrow"].tail_length = 0.5
        cfg.markers["arrow"].head_radius = 0.02
        cfg.markers["arrow"].head_length = 0.1
        cfg.markers["arrow"].visual_material = PreviewSurfaceCfg(
            emissive_color=(0.0, 1.0, 0.0)
        )
        self._marker_target_linvel = VisualizationMarkers(cfg)
        cfg = ARROW_CFG.copy().replace(  # type: ignore
            prim_path="/Visuals/command/robot_linvel"
        )
        cfg.markers["arrow"].tail_radius = 0.01
        cfg.markers["arrow"].tail_length = 0.5
        cfg.markers["arrow"].head_radius = 0.02
        cfg.markers["arrow"].head_length = 0.1
        cfg.markers["arrow"].visual_material = PreviewSurfaceCfg(
            emissive_color=(0.2, 0.8, 0.2)
        )
        self._marker_robot_linvel = VisualizationMarkers(cfg)

        ## Angular velocity
        cfg = ARROW_CFG.copy().replace(  # type: ignore
            prim_path="/Visuals/command/target_angvel"
        )
        cfg.markers["arrow"].tail_length = 0.0
        cfg.markers["arrow"].tail_radius = 0.0
        cfg.markers["arrow"].head_radius = 0.025
        cfg.markers["arrow"].head_length = 0.15
        cfg.markers["arrow"].visual_material = PreviewSurfaceCfg(
            emissive_color=(0.2, 0.2, 0.8)
        )
        self._marker_target_angvel = VisualizationMarkers(cfg)
        cfg = ARROW_CFG.copy().replace(  # type: ignore
            prim_path="/Visuals/command/robot_angvel"
        )
        cfg.markers["arrow"].tail_length = 0.0
        cfg.markers["arrow"].tail_radius = 0.0
        cfg.markers["arrow"].head_radius = 0.025
        cfg.markers["arrow"].head_length = 0.15
        cfg.markers["arrow"].visual_material = PreviewSurfaceCfg(
            emissive_color=(0.2, 0.2, 0.8)
        )
        self._marker_robot_angvel = VisualizationMarkers(cfg)

    def _update_visualization_markers(self):
        MARKER_OFFSET_Z_LINVEL = 0.2
        MARKER_OFFSET_Z_ANGVEL = 0.175

        ## Common
        robot_pos_w = self._robot.data.root_link_pos_w
        marker_pos = torch.zeros(
            (self.cfg.scene.num_envs, 3), dtype=torch.float32, device=self.device
        )
        marker_orientation = torch.zeros(
            (self.cfg.scene.num_envs, 4), dtype=torch.float32, device=self.device
        )
        marker_scale = torch.ones(
            (self.cfg.scene.num_envs, 3), dtype=torch.float32, device=self.device
        )
        marker_pos[:, :2] = robot_pos_w[:, :2]

        ## Target linear velocity
        marker_pos[:, 2] = robot_pos_w[:, 2] + MARKER_OFFSET_Z_LINVEL
        marker_heading = self._robot.data.heading_w + torch.atan2(
            self._command[:, 1], self._command[:, 0]
        )
        marker_orientation[:, 0] = torch.cos(marker_heading * 0.5)
        marker_orientation[:, 3] = torch.sin(marker_heading * 0.5)
        marker_scale[:, 0] = torch.norm(
            torch.stack(
                (self._command[:, 0], self._command[:, 1]),
                dim=-1,
            ),
            dim=-1,
        )
        self._marker_target_linvel.visualize(
            marker_pos, marker_orientation, marker_scale
        )

        ## Robot linear velocity
        marker_heading = self._robot.data.heading_w + torch.atan2(
            self._robot.data.root_lin_vel_b[:, 1],
            self._robot.data.root_lin_vel_b[:, 0],
        )
        marker_orientation[:, 0] = torch.cos(marker_heading * 0.5)
        marker_orientation[:, 3] = torch.sin(marker_heading * 0.5)
        marker_scale[:, 0] = torch.norm(self._robot.data.root_lin_vel_b[:, :2], dim=-1)
        self._marker_robot_linvel.visualize(
            marker_pos, marker_orientation, marker_scale
        )

        ## Target angular velocity
        _target_angvel_abs = self._command[:, 2].abs()
        normalization_fac = torch.where(
            self._command[:, 2] != 0.0,
            (torch.pi / 2.0) / self._command[:, 2].abs(),
            torch.ones_like(self._command[:, 2]),
        ).clamp(max=1.0)
        marker_pos[:, 2] = robot_pos_w[:, 2] + MARKER_OFFSET_Z_ANGVEL
        marker_heading = (
            self._robot.data.heading_w + normalization_fac * self._command[:, 2]
        )
        marker_orientation[:, 0] = torch.cos(marker_heading * 0.5)
        marker_orientation[:, 3] = torch.sin(marker_heading * 0.5)
        marker_scale[:, 0] = 1.0
        self._marker_target_angvel.visualize(
            marker_pos, marker_orientation, marker_scale
        )

        ## Robot angular velocity
        marker_heading = (
            self._robot.data.heading_w
            + normalization_fac * self._robot.data.root_ang_vel_w[:, -1]
        )
        marker_orientation[:, 0] = torch.cos(marker_heading * 0.5)
        marker_orientation[:, 3] = torch.sin(marker_heading * 0.5)
        marker_scale[:, 0] = 1.0
        self._marker_robot_angvel.visualize(
            marker_pos, marker_orientation, marker_scale
        )

    def extract_step_return(self) -> StepReturn:
        if self.cfg.command_vis or self.cfg.debug_vis:
            self._update_visualization_markers()

        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Root
            tf_quat_robot=self._robot.data.root_quat_w,
            vel_lin_robot=self._robot.data.root_lin_vel_b,
            vel_ang_robot=self._robot.data.root_ang_vel_b,
            # IMU
            imu_lin_acc=self._imu_robot.data.lin_acc_b,
            imu_ang_vel=self._imu_robot.data.ang_vel_b,
            ## Command
            command=self._command,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Root
    tf_quat_robot: torch.Tensor,
    vel_lin_robot: torch.Tensor,
    vel_ang_robot: torch.Tensor,
    # IMU
    imu_lin_acc: torch.Tensor,
    imu_ang_vel: torch.Tensor,
    ## Command
    command: torch.Tensor,
) -> StepReturn:
    num_envs = episode_length.size(0)
    # dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Root
    tf_rotmat_robot = matrix_from_quat(tf_quat_robot)
    tf_rot6d_robot = rotmat_to_rot6d(tf_rotmat_robot)

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Reward: Command tracking (linear)
    WEIGHT_CMD_LIN_VEL_XY = 3.0
    EXP_STD_CMD_LIN_VEL_XY = 0.5
    reward_cmd_lin_vel_xy = WEIGHT_CMD_LIN_VEL_XY * torch.exp(
        -torch.sum(torch.square(command[:, :2] - vel_lin_robot[:, :2]), dim=1)
        / EXP_STD_CMD_LIN_VEL_XY
    )

    # Reward: Command tracking (angular)
    WEIGHT_CMD_ANG_VEL_Z = 1.5
    EXP_STD_CMD_ANG_VEL_Z = 0.25
    reward_cmd_ang_vel_z = WEIGHT_CMD_ANG_VEL_Z * torch.exp(
        -torch.square(command[:, 2] - vel_ang_robot[:, 2]) / EXP_STD_CMD_ANG_VEL_Z
    )

    # Penalty: Minimize non-command motion (linear)
    WEIGHT_UNDESIRED_LIN_VEL_Z = -0.5
    penalty_undesired_lin_vel_z = WEIGHT_UNDESIRED_LIN_VEL_Z * torch.square(
        vel_lin_robot[:, 2]
    )

    # Penalty: Minimize non-command motion (angular)
    WEIGHT_UNDESIRED_ANG_VEL_XY = -0.1
    penalty_undesired_ang_vel_xy = WEIGHT_UNDESIRED_ANG_VEL_XY * torch.sum(
        torch.square(vel_ang_robot[:, :2]), dim=-1
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "tf_rot6d_robot": tf_rot6d_robot,
                "vel_lin_robot": vel_lin_robot,
                "vel_ang_robot": vel_ang_robot,
            },
            "proprio": {
                "imu_lin_acc": imu_lin_acc,
                "imu_ang_vel": imu_ang_vel,
            },
            "command": {
                "cmd_vel": command,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "reward_cmd_lin_vel_xy": reward_cmd_lin_vel_xy,
            "reward_cmd_ang_vel_z": reward_cmd_ang_vel_z,
            "penalty_undesired_lin_vel_z": penalty_undesired_lin_vel_z,
            "penalty_undesired_ang_vel_xy": penalty_undesired_ang_vel_xy,
        },
        termination,
        truncation,
    )



from __future__ import annotations

from functools import cache, cached_property
from itertools import chain
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Mapping,
    Sequence,
    Set,
    Tuple,
    Type,
)

from pydantic import BaseModel, PositiveFloat
from simforge import BlGeometry, BlModel, BlShader, TexResConfig

from srb.core.asset import AssetBaseCfg
from srb.core.asset.asset_type import AssetType
from srb.core.asset.asset_variant import AssetVariant
from srb.core.domain import Domain
from srb.core.sim import MultiAssetSpawnerCfg, ShapeCfg, SimforgeAssetCfg, SpawnerCfg
from srb.utils import logging
from srb.utils.str import convert_to_snake_case


class Asset(BaseModel):
    ## Scenario
    DOMAINS: ClassVar[Sequence[Domain]] = ()

    ## Model
    asset_cfg: AssetBaseCfg

    ## Attributes forwarded to the spawner
    SPAWNER_ATTRIBUTES: ClassVar[Sequence[str]] = (
        "size",
        "scale",
        "texture_resolution",
    )
    size: Tuple[PositiveFloat, PositiveFloat] | None = None
    scale: Tuple[PositiveFloat, PositiveFloat, PositiveFloat] | None = None
    texture_resolution: TexResConfig | None = None

    @classmethod
    @cache
    def name(cls) -> str:
        return convert_to_snake_case(cls.__name__)

    def __new__(cls, *args, **kwargs):
        if cls in (
            Asset,
            *AssetRegistry.base_types.keys(),
            *AssetRegistry.meta_types,
        ):
            raise TypeError(f"Cannot instantiate abstract class {cls.__name__}")
        return super().__new__(cls)

    def __init_subclass__(
        cls,
        asset_entrypoint: AssetType | None = None,
        asset_metaclass: bool = False,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)
        if asset_entrypoint is not None:
            assert isinstance(asset_entrypoint, AssetType), (
                f"Class '{cls.__name__}' is marked as an asset entrypoint, but '{asset_entrypoint}' is not a valid {AssetType}"
            )
            assert asset_entrypoint not in AssetRegistry.base_types.keys(), (
                f"Class '{cls.__name__}' is marked as '{asset_entrypoint}' asset entrypoint, but it was already marked by '{AssetRegistry.base_types[asset_entrypoint].__name__}'"
            )
            AssetRegistry.base_types[asset_entrypoint] = cls
        elif asset_metaclass:
            AssetRegistry.meta_types.append(cls)
        else:
            for asset_type, base in AssetRegistry.base_types.items():
                if issubclass(cls, base):
                    if asset_type not in AssetRegistry.registry.keys():
                        AssetRegistry.registry[asset_type] = []
                    else:
                        assert cls.name() not in (
                            asset.name() for asset in AssetRegistry.registry[asset_type]
                        ), (
                            f"Cannot register multiple assets with an identical name: '{cls.__module__}:{cls.__name__}' already exists as '{next(asset for asset in AssetRegistry.registry[asset_type] if cls.name() == asset.name()).__module__}:{cls.__name__}'"
                        )
                    AssetRegistry.registry[asset_type].append(cls)
                    break

    @cached_property
    def asset_type(self) -> AssetType:
        for asset_type, base in AssetRegistry.base_types.items():
            if isinstance(self, base):
                return asset_type
        raise ValueError(f"Class '{self.__class__.__name__}' has unknown asset type")

    @classmethod
    def asset_registry(cls) -> Sequence[Type[Asset]]:
        return list(AssetRegistry.values_inner())

    @cached_property
    def asset_variant(self) -> AssetVariant | None:
        if not isinstance(self.asset_cfg, AssetBaseCfg) or self.asset_cfg.spawn is None:
            return None

        def __get_variant(spawner: SpawnerCfg) -> AssetVariant:
            if isinstance(spawner, SimforgeAssetCfg):
                return AssetVariant.PROCEDURAL
            elif isinstance(spawner, ShapeCfg):
                return AssetVariant.PRIMITIVE
            else:
                return AssetVariant.DATASET

        if isinstance(self.asset_cfg.spawn, MultiAssetSpawnerCfg):
            assert len(self.asset_cfg.spawn.assets_cfg) > 0
            variants = set(
                __get_variant(spawner) for spawner in self.asset_cfg.spawn.assets_cfg
            )
            if AssetVariant.PROCEDURAL in variants:
                return AssetVariant.PROCEDURAL
            elif AssetVariant.DATASET in variants:
                return AssetVariant.DATASET
            else:
                return AssetVariant.PRIMITIVE
        else:
            return __get_variant(self.asset_cfg.spawn)

    def is_variant(self, variant: AssetVariant) -> bool:
        return variant is self.asset_variant

    @property
    def _model_values(self) -> Mapping[str, Any]:
        return {
            k: v
            for k, v in {
                k: getattr(self, k)
                for k in chain(
                    self.__class__.model_fields.keys(),
                    self.model_computed_fields.keys(),
                )
            }.items()
            if v is not None
        }

    def model_post_init(self, __context):
        super().model_post_init(__context)

        ## Forward specified attributes to the spawner
        if isinstance(self.asset_cfg, AssetBaseCfg):
            assert self.asset_cfg.spawn is not None
            Asset.__set_spawner_attributes(
                self.asset_cfg.spawn, self._spawner_attributes
            )

    @property
    def _spawner_attributes(self) -> Mapping[str, Any]:
        return {
            k: v
            for k, v in self._model_values.items()
            if k in self._get_combined_spawner_attribute_keys()
        }

    def _get_combined_spawner_attribute_keys(self) -> Set[str]:
        all_attributes = set()
        for cls in self.__class__.__mro__:
            if hasattr(cls, "SPAWNER_ATTRIBUTES"):
                all_attributes.update(cls.SPAWNER_ATTRIBUTES)
        return all_attributes

    @staticmethod
    def __set_spawner_attributes(
        spawner: SpawnerCfg, spawner_attributes: Mapping[str, Any]
    ):
        if len(spawner_attributes) == 0:
            return

        if isinstance(spawner, SimforgeAssetCfg):
            for asset in spawner.assets:
                match asset:
                    case _geo if isinstance(asset, BlGeometry):
                        for k, v in spawner_attributes.items():
                            _set = False
                            for op in asset.ops:
                                if hasattr(op, k) and isinstance(
                                    getattr(op, k), type(v)
                                ):
                                    setattr(op, k, v)
                                    _set = True
                                    logging.debug(
                                        f'Updated input "{k}" to "{v}" for {BlGeometry.__name__} operation "{op.__class__.__name__}" of "{asset.__class__.__name__}"'
                                    )
                            if not _set:
                                logging.trace(
                                    f'Input "{k}" of type "{type(k)}" not updated for "{asset.__class__.__name__}"'
                                )
                    case _model if isinstance(asset, BlModel):
                        for k, v in spawner_attributes.items():
                            _set = False
                            for op in asset.geo.ops:
                                if hasattr(op, k) and isinstance(
                                    getattr(op, k), type(v)
                                ):
                                    setattr(op, k, v)
                                    _set = True
                                    logging.debug(
                                        f'Updated input "{k}" to "{v}" for {BlGeometry.__name__} operation "{op.__class__.__name__}" of "{asset.__class__.__name__}/{asset.geo.__class__.__name__}"'
                                    )
                            if (
                                asset.mat is not None
                                and hasattr(asset.mat.shader, k)
                                and isinstance(getattr(asset.mat.shader, k), type(v))
                            ):
                                setattr(asset.mat.shader, k, v)
                                _set = True
                                logging.debug(
                                    f'Updated input "{k}" to "{v}" for "{asset.mat.shader.__class__.__name__}" {BlShader.__name__} of "{asset.__class__.__name__}/{asset.mat.__class__.__name__}"'
                                )
                            if hasattr(asset, k) and isinstance(
                                getattr(asset, k), type(v)
                            ):
                                setattr(asset, k, v)
                                _set = True
                                logging.debug(
                                    f'Updated input "{k}" to "{v}" for "{asset.__class__.__name__}"'
                                )
                            if not _set:
                                logging.trace(
                                    f'Input "{k}" of type "{type(k)}" not updated for "{asset.__class__.__name__}"'
                                )
                    case _:
                        logging.warning(
                            f"SimForge asset of type '{type(asset)}' is not supported for input updates"
                        )
        elif isinstance(spawner, MultiAssetSpawnerCfg):
            for subspawner in spawner.assets_cfg:
                Asset.__set_spawner_attributes(subspawner, spawner_attributes)
            for k, v in spawner_attributes.items():
                _set = False
                if hasattr(spawner, k) and isinstance(getattr(spawner, k), type(v)):
                    setattr(spawner, k, v)
                if not _set:
                    logging.trace(
                        f'Input "{k}" of type "{type(k)}" not updated for "{spawner.__class__.__name__}"'
                    )
        else:
            for k, v in spawner_attributes.items():
                _set = False
                if hasattr(spawner, k) and isinstance(getattr(spawner, k), type(v)):
                    setattr(spawner, k, v)
                    _set = True
                    logging.debug(
                        f'Updated input "{k}" to "{v}" for "{spawner.__class__.__name__}"'
                    )
                if not _set:
                    logging.trace(
                        f'Input "{k}" of type "{type(k)}" not updated for "{spawner.__class__.__name__}"'
                    )


class AssetRegistry:
    registry: ClassVar[Dict[AssetType, List[Type[Asset]]]] = {}
    base_types: ClassVar[Dict[AssetType, Type[Asset]]] = {}
    meta_types: ClassVar[List[Type[Asset]]] = []

    @classmethod
    def keys(cls) -> Iterable[AssetType]:
        return cls.registry.keys()

    @classmethod
    def items(cls) -> Iterable[Tuple[AssetType, Sequence[Type[Asset]]]]:
        return cls.registry.items()

    @classmethod
    def values(cls) -> Iterable[Iterable[Type[Asset]]]:
        return cls.registry.values()

    @classmethod
    def values_inner(cls) -> Iterable[Type[Asset]]:
        return (asset for assets in cls.registry.values() for asset in assets)

    @classmethod
    def n_assets(cls) -> int:
        return sum(len(assets) for assets in cls.registry.values())

    @classmethod
    def registered_modules(cls) -> Iterable[str]:
        return {
            asset.__module__ for assets in cls.registry.values() for asset in assets
        }

    @classmethod
    def registered_packages(cls) -> Iterable[str]:
        return {module.split(".", maxsplit=1)[0] for module in cls.registered_modules()}

    @classmethod
    def get_by_name(cls, name: str) -> Type[Asset] | None:
        for asset in cls.values_inner():
            if asset.name() == name:
                return asset
        return None



from srb.core.action import (  # noqa: F401
    ActionGroup,
    DifferentialIKControllerCfg,
    DifferentialInverseKinematicsActionCfg,
    InverseKinematicsActionGroup,
    OperationalSpaceControlActionGroup,
    OperationalSpaceControllerActionCfg,
    OperationalSpaceControllerCfg,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, SerialManipulator, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class Canadarm3(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/canadarm3_large",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("manipulator")
            .joinpath("canadarm3_large.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "canadarm3_large_joint_1": deg_to_rad(50.0),
                "canadarm3_large_joint_2": 0.0,
                "canadarm3_large_joint_3": deg_to_rad(55.0),
                "canadarm3_large_joint_4": deg_to_rad(75.0),
                "canadarm3_large_joint_5": deg_to_rad(-30.0),
                "canadarm3_large_joint_6": 0.0,
                "canadarm3_large_joint_7": 0.0,
            },
        ),
        actuators={
            "joints": ImplicitActuatorCfg(
                joint_names_expr=["canadarm3_large_joint_[1-7]"],
                effort_limit=2500.0,
                velocity_limit=5.0,
                stiffness=40000.0,
                damping=25000.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=["canadarm3_large_joint_[1-7]"],
            base_name="canadarm3_large_0",
            body_name="canadarm3_large_7",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(
                pos=(0.0, 0.0, -0.45),
            ),
        ),
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="canadarm3_large_0")
    frame_flange: Frame = Frame(
        prim_relpath="canadarm3_large_7",
        offset=Transform(
            pos=(0.0, 0.0, -0.44),
            rot=rpy_to_quat(0.0, 180.0, 0.0),
        ),
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="canadarm3_large_0/camera_base",
        offset=Transform(
            pos=(0.06, 0.0, 0.15),
            rot=rpy_to_quat(0.0, -10.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="canadarm3_large_7/camera_wrist",
        offset=Transform(
            pos=(0.0, 0.0, -0.45),
            rot=rpy_to_quat(0.0, 90.0, 180.0),
        ),
    )



from dataclasses import MISSING
from typing import TYPE_CHECKING, Dict, Sequence, Type

import torch

from srb.core.manager import ActionTerm, ActionTermCfg
from srb.utils.cfg import configclass
from srb.utils.math import euler_xyz_from_quat, quat_from_euler_xyz

if TYPE_CHECKING:
    from srb._typing import AnyEnv
    from srb.core.asset import Articulation


class MulticopterBodyAccelerationAction(ActionTerm):
    cfg: "MulticopterBodyAccelerationActionCfg"
    _env: "AnyEnv"
    _asset: "Articulation"

    def __init__(self, cfg: "MulticopterBodyAccelerationActionCfg", env: "AnyEnv"):
        super().__init__(
            cfg,
            env,  # type: ignore
        )

        self._body_index = self._asset.find_bodies(self.cfg.frame_base)[0]
        self._rotor_indices, rotor_names = self._asset.find_joints(
            self.cfg.regex_rotor_joints
        )

        if isinstance(self.cfg.nominal_rpm, Dict):
            self._nominal_rpm = torch.tensor(
                [self.cfg.nominal_rpm[name] for name in rotor_names],
                device=self.device,
                dtype=torch.float32,
            ).repeat(self._asset.num_instances, 1)
        else:
            self._nominal_rpm = torch.tensor(
                [self.cfg.nominal_rpm] * len(rotor_names),
                device=self.device,
                dtype=torch.float32,
            ).repeat(self._asset.num_instances, 1)

    @property
    def action_dim(self) -> int:
        return 4

    @property
    def raw_actions(self) -> torch.Tensor:
        return self._raw_actions

    @property
    def processed_actions(self) -> torch.Tensor:
        return self._processed_actions

    def process_actions(self, actions):
        self._raw_actions = actions
        self._processed_actions = self.raw_actions * self.cfg.scale

    def apply_actions(self):
        self._asset.set_joint_velocity_target(
            self._nominal_rpm, joint_ids=self._rotor_indices
        )

        current_velocity = self._asset._data.body_vel_w[:, self._body_index].squeeze(1)
        current_yaw = euler_xyz_from_quat(self._asset._data.root_quat_w)[2]

        applied_velocity_lin = (
            self.cfg.controller_damping * current_velocity[:, :3]
            + (1 - self.cfg.controller_damping) * self.processed_actions[:, :3]
        )
        applied_velocity_rot_yaw = (
            self.cfg.controller_damping * current_velocity[:, 5]
            + (1 - self.cfg.controller_damping) * self.processed_actions[:, 3]
        )
        applied_velocities = torch.cat(
            (
                applied_velocity_lin,
                torch.zeros_like(applied_velocity_rot_yaw)[:, None],
                torch.zeros_like(applied_velocity_rot_yaw)[:, None],
                applied_velocity_rot_yaw[:, None],
            ),
            dim=1,
        )
        if self.cfg.noise > 0:
            applied_velocities += self.cfg.noise * torch.randn_like(applied_velocities)
        self._asset.write_root_velocity_to_sim(applied_velocities)

        target_tilt = (
            torch.atan2(applied_velocity_lin[:, 1], applied_velocity_lin[:, 0])
            - current_yaw
        )
        tilt_factor = self.cfg.tilt_magnitude * torch.norm(
            current_velocity[:, :2], dim=1
        )
        target_roll = -torch.sin(target_tilt) * tilt_factor
        target_pitch = torch.cos(target_tilt) * tilt_factor
        self._asset.write_root_pose_to_sim(
            torch.cat(
                (
                    self._asset._data.root_pos_w,
                    quat_from_euler_xyz(target_roll, target_pitch, current_yaw),
                ),
                dim=1,
            )
        )

    def reset(self, env_ids: Sequence[int] | None = None) -> None:
        pass


@configclass
class MulticopterBodyAccelerationActionCfg(ActionTermCfg):
    class_type: Type = MulticopterBodyAccelerationAction  # type: ignore

    frame_base: str = MISSING  # type: ignore
    regex_rotor_joints: str = MISSING  # type: ignore

    nominal_rpm: float | Dict[str, float] = 1000.0
    tilt_magnitude: float = 1.0
    controller_damping: float = 0.98

    scale: float = 1.0
    noise: float = 0.0025



from typing import TYPE_CHECKING, Type

import torch

from srb.core.manager import ActionTerm, ActionTermCfg
from srb.utils.cfg import configclass

if TYPE_CHECKING:
    from srb._typing import AnyEnv
    from srb.core.asset import Articulation, RigidObject


class BodyAccelerationAction(ActionTerm):
    cfg: "BodyAccelerationActionCfg"
    _env: "AnyEnv"
    _asset: "Articulation | RigidObject"

    @property
    def action_dim(self) -> int:
        return 6

    @property
    def raw_actions(self) -> torch.Tensor:
        return self._raw_actions

    @property
    def processed_actions(self) -> torch.Tensor:
        return self._processed_actions

    def process_actions(self, actions):
        self._raw_actions = actions
        self._processed_actions = self.raw_actions * self.cfg.scale

    def apply_actions(self):
        applied_velocities = (
            self._asset._data.body_vel_w[:, 0].squeeze(1) + self.processed_actions
        )
        if self.cfg.relative:
            applied_velocities += self._asset._data.body_acc_w[:, 0].squeeze(1)

        self._asset.write_root_velocity_to_sim(applied_velocities)


@configclass
class BodyAccelerationActionCfg(ActionTermCfg):
    class_type: Type = BodyAccelerationAction

    relative: bool = False
    scale: float = 1.0



from isaaclab.devices import *  # noqa: F403

from .combined import CombinedTeleopInterface  # noqa: F401
from .keyboard import (  # noqa: F401
    EventKeyboardTeleopInterface,
    KeyboardTeleopInterface,
)
from .spacemouse import SpacemouseTeleopInterface  # noqa: F401



from srb.core.action import (
    ActionGroup,
    JointPositionActionCfg,
    JointPositionActionGroup,
)
from srb.core.actuator import ActuatorNetMLPCfg, DCMotorCfg
from srb.core.asset import ArticulationCfg, Frame, LeggedRobot, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    MultiAssetSpawnerCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import rpy_to_quat
from srb.utils.nucleus import ISAACLAB_NUCLEUS_DIR


class UnitreeA1(LeggedRobot):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/unitree_a1",
        spawn=UsdFileCfg(
            usd_path=f"{ISAACLAB_NUCLEUS_DIR}/Robots/Unitree/A1/a1.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                retain_accelerations=False,
                linear_damping=0.0,
                angular_damping=0.0,
                max_linear_velocity=1000.0,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.42),
            joint_pos={
                ".*L_hip_joint": 0.1,
                ".*R_hip_joint": -0.1,
                "F[L,R]_thigh_joint": 0.8,
                "R[L,R]_thigh_joint": 1.0,
                ".*_calf_joint": -1.5,
            },
            joint_vel={".*": 0.0},
        ),
        actuators={
            "base_legs": DCMotorCfg(
                joint_names_expr=[".*_hip_joint", ".*_thigh_joint", ".*_calf_joint"],
                effort_limit=33.5,
                saturation_effort=33.5,
                velocity_limit=21.0,
                stiffness=25.0,
                damping=0.5,
                friction=0.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="trunk")
    frame_payload_mount: Frame = Frame(
        prim_relpath="trunk",
        offset=Transform(
            pos=(-0.1, 0.0, 0.06),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="trunk",
        offset=Transform(
            pos=(0.15, 0.0, 0.06),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="trunk/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = ".*foot"


class UnitreeGo1(LeggedRobot):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/unitree_go1",
        spawn=UsdFileCfg(
            usd_path=f"{ISAACLAB_NUCLEUS_DIR}/Robots/Unitree/Go1/go1.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                retain_accelerations=False,
                linear_damping=0.0,
                angular_damping=0.0,
                max_linear_velocity=1000.0,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.4),
            joint_pos={
                ".*L_hip_joint": 0.1,
                ".*R_hip_joint": -0.1,
                "F[L,R]_thigh_joint": 0.8,
                "R[L,R]_thigh_joint": 1.0,
                ".*_calf_joint": -1.5,
            },
            joint_vel={".*": 0.0},
        ),
        actuators={
            "base_legs": ActuatorNetMLPCfg(
                joint_names_expr=[".*_hip_joint", ".*_thigh_joint", ".*_calf_joint"],
                network_file=f"{ISAACLAB_NUCLEUS_DIR}/ActuatorNets/Unitree/unitree_go1.pt",
                pos_scale=-1.0,
                vel_scale=1.0,
                torque_scale=1.0,
                input_order="pos_vel",
                input_idx=[0, 1, 2],
                effort_limit=23.7,
                velocity_limit=30.0,
                saturation_effort=23.7,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="trunk")
    frame_payload_mount: Frame = Frame(
        prim_relpath="trunk",
        offset=Transform(
            pos=(-0.1, 0.0, 0.06),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="trunk",
        offset=Transform(
            pos=(0.15, 0.0, 0.06),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="trunk/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = ".*foot"


class UnitreeGo2(LeggedRobot):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/unitree_go2",
        spawn=UsdFileCfg(
            usd_path=f"{ISAACLAB_NUCLEUS_DIR}/Robots/Unitree/Go2/go2.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                retain_accelerations=False,
                linear_damping=0.0,
                angular_damping=0.0,
                max_linear_velocity=1000.0,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.42),
            joint_pos={
                ".*L_hip_joint": 0.1,
                ".*R_hip_joint": -0.1,
                "F[L,R]_thigh_joint": 0.8,
                "R[L,R]_thigh_joint": 1.0,
                ".*_calf_joint": -1.5,
            },
            joint_vel={".*": 0.0},
        ),
        actuators={
            "base_legs": DCMotorCfg(
                joint_names_expr=[".*_hip_joint", ".*_thigh_joint", ".*_calf_joint"],
                effort_limit=33.5,
                saturation_effort=33.5,
                velocity_limit=21.0,
                stiffness=25.0,
                damping=0.5,
                friction=0.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(-0.1, 0.0, 0.06),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.15, 0.0, 0.06),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="base/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = ".*foot"


class RandomUnitreeQuadruped(LeggedRobot):
    asset_cfg: ArticulationCfg = UnitreeGo2().asset_cfg.copy()  # type: ignore
    asset_cfg.prim_path = "{ENV_REGEX_NS}/anymal"
    asset_cfg.spawn = MultiAssetSpawnerCfg(
        random_choice=False,
        assets_cfg=(
            UnitreeA1().asset_cfg.spawn,  # type: ignore
            UnitreeGo1().asset_cfg.spawn,  # type: ignore
        ),
        activate_contact_sensors=True,
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="trunk")
    frame_payload_mount: Frame = Frame(
        prim_relpath="trunk",
        offset=Transform(
            pos=(-0.1, 0.0, 0.06),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="trunk",
        offset=Transform(
            pos=(0.15, 0.0, 0.06),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="trunk/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = ".*foot"



import collections
from functools import partial as bind

import elements
import embodied
import numpy
from isaacsim.simulation_app import SimulationApp

from srb.integrations.dreamer.driver import DriverParallelEnv


def train(
    make_agent,
    make_replay,
    make_env,
    make_stream,
    make_logger,
    args,
    sim_app: SimulationApp,
):
    agent = make_agent()
    replay = make_replay()
    logger = make_logger()

    logdir = elements.Path(args.logdir)
    step = logger.step
    usage = elements.Usage(**args.usage)
    train_agg = elements.Agg()
    epstats = elements.Agg()
    episodes = collections.defaultdict(elements.Agg)
    policy_fps = elements.FPS()
    train_fps = elements.FPS()

    batch_steps = args.batch_size * args.batch_length
    should_train = elements.when.Ratio(args.train_ratio / batch_steps)
    should_log = embodied.LocalClock(args.log_every)
    should_report = embodied.LocalClock(args.report_every)
    should_save = embodied.LocalClock(args.save_every)

    @elements.timer.section("logfn")
    def logfn(tran, worker):
        episode = episodes[worker]
        tran["is_first"] and episode.reset()  # type: ignore
        episode.add("score", tran["reward"], agg="sum")
        episode.add("length", 1, agg="sum")
        episode.add("rewards", tran["reward"], agg="stack")
        for key, value in tran.items():
            if value.dtype == numpy.uint8 and value.ndim == 3:
                if worker == 0:
                    if value.shape[-1] >= 3:
                        episode.add(f"policy_{key}_ch0-2", value[..., :3], agg="stack")
                        for i in range(3, value.shape[-1]):
                            episode.add(
                                f"policy_{key}_ch{i}",
                                value[..., i : i + 1],
                                agg="stack",
                            )
                    else:
                        for i in range(value.shape[-1]):
                            episode.add(
                                f"policy_{key}_ch{i}",
                                value[..., i : i + 1],
                                agg="stack",
                            )
            elif key.startswith("log/"):
                assert value.ndim == 0, (key, value.shape, value.dtype)
                episode.add(key, value, agg=("avg", "max", "sum"))  # type: ignore
        if tran["is_last"]:
            result = episode.result()
            logger.add(
                {
                    "score": result.pop("score"),
                    "length": result.pop("length"),
                },
                prefix="episode",
            )
            rew = result.pop("rewards")
            if len(rew) > 1:
                result["reward_rate"] = (numpy.abs(rew[1:] - rew[:-1]) >= 0.01).mean()
            epstats.add(result)

    ########### ONLY THIS IS CHANGED (+1 fix below) ############
    # fns = [bind(make_env, i) for i in range(args.envs)]
    # driver = embodied.Driver(fns, parallel=not args.debug)
    driver = DriverParallelEnv(env=make_env(), num_envs=args.envs)
    ########### ONLY THIS IS CHANGED (+1 fix below) ############
    driver.on_step(lambda tran, _: step.increment())
    driver.on_step(lambda tran, _: policy_fps.step())
    driver.on_step(replay.add)
    driver.on_step(logfn)

    stream_train = iter(agent.stream(make_stream(replay, "train")))
    stream_report = iter(agent.stream(make_stream(replay, "report")))

    carry_train = [agent.init_train(args.batch_size)]
    carry_report = agent.init_report(args.batch_size)

    def trainfn(tran, worker):
        if len(replay) < args.batch_size * args.batch_length:
            return
        for _ in range(should_train(step)):
            with elements.timer.section("stream_next"):
                batch = next(stream_train)
            carry_train[0], outs, mets = agent.train(carry_train[0], batch)
            train_fps.step(batch_steps)
            if "replay" in outs:
                replay.update(outs["replay"])
            train_agg.add(mets, prefix="train")

    driver.on_step(trainfn)

    cp = elements.Checkpoint(logdir / "ckpt")
    cp.step = step
    cp.agent = agent
    cp.replay = replay
    if args.from_checkpoint:
        ########### Fix below ############
        # elements.checkpoint.load(
        #     args.from_checkpoint,
        #     dict(agent=bind(agent.load, regex=args.from_checkpoint_regex)),
        # )
        elements.checkpoint.load(args.from_checkpoint, dict(agent=bind(agent.load)))
        ########### Fix below ############
    cp.load_or_save()

    print("Start training loop")
    policy = lambda *args: agent.policy(*args, mode="train")  # noqa: E731
    driver.reset(agent.init_policy)
    while step < args.steps:
        ########### Sim extra (optional) ############
        if not sim_app.is_running():
            break
        ########### Sim extra (optional) ############
        driver(policy, steps=10)

        if should_report(step) and len(replay):
            agg = elements.Agg()
            for _ in range(args.consec_report * args.report_batches):
                carry_report, mets = agent.report(carry_report, next(stream_report))
                agg.add(mets)
            logger.add(agg.result(), prefix="report")

        if should_log(step):
            logger.add(train_agg.result())
            logger.add(epstats.result(), prefix="epstats")
            logger.add(replay.stats(), prefix="replay")
            logger.add(usage.stats(), prefix="usage")
            logger.add({"fps/policy": policy_fps.result()})
            logger.add({"fps/train": train_fps.result()})
            logger.add({"timer": elements.timer.stats()["summary"]})
            logger.write()

        if should_save(step):
            cp.save()

    logger.close()



from dataclasses import MISSING

import torch

from srb.core.action import (
    BinaryJointPositionActionCfg,
    BinaryJointVelocityActionCfg,
    EMAJointPositionToLimitsActionCfg,
    JointEffortActionCfg,
    JointPositionActionCfg,
    JointPositionToLimitsActionCfg,
    JointVelocityActionCfg,
    RelativeJointPositionActionCfg,
)
from srb.core.action.action_group import ActionGroup
from srb.utils.cfg import configclass


@configclass
class JointPositionActionGroup(ActionGroup):
    joint_pos: JointPositionActionCfg = JointPositionActionCfg(
        asset_name="robot", joint_names=[".*"]
    )


@configclass
class JointPositionRelativeActionGroup(ActionGroup):
    joint_pos: RelativeJointPositionActionCfg = RelativeJointPositionActionCfg(
        asset_name="robot", joint_names=[".*"]
    )


@configclass
class JointPositionBoundedActionGroup(ActionGroup):
    joint_pos: JointPositionToLimitsActionCfg = JointPositionToLimitsActionCfg(
        asset_name="robot", joint_names=[".*"]
    )


@configclass
class JointPositionBoundedEMAActionGroup(ActionGroup):
    joint_pos: EMAJointPositionToLimitsActionCfg = EMAJointPositionToLimitsActionCfg(
        asset_name="robot", joint_names=[".*"]
    )


@configclass
class JointPositionBinaryActionGroup(ActionGroup):
    joint_pos: BinaryJointPositionActionCfg = MISSING  # type: ignore

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        return torch.Tensor((-1.0 if event else 1.0,)).to(device=twist.device)


@configclass
class JointVelocityActionGroup(ActionGroup):
    joint_vel: JointVelocityActionCfg = JointVelocityActionCfg(
        asset_name="robot", joint_names=[".*"]
    )


@configclass
class JointVelocityBinaryActionGroup(ActionGroup):
    joint_vel: BinaryJointVelocityActionCfg = MISSING  # type: ignore

    def map_cmd_to_action(self, twist: torch.Tensor, event: bool) -> torch.Tensor:
        return torch.Tensor((-1.0 if event else 1.0,)).to(device=twist.device)


@configclass
class JointEffortActionGroup(ActionGroup):
    joint_eff: JointEffortActionCfg = JointEffortActionCfg(
        asset_name="robot", joint_names=[".*"]
    )



try:
    import rclpy
except ImportError:
    from srb.utils.ros import enable_ros2_bridge

    enable_ros2_bridge()

import os
import threading
from collections.abc import Callable

import numpy
import rclpy
import torch
from geometry_msgs.msg import Twist, Vector3
from rclpy.executors import MultiThreadedExecutor
from rclpy.node import Node
from std_msgs.msg import Bool, Float64

from srb.interfaces.teleop import DeviceBase


class HapticROSTeleopInterface(DeviceBase):
    def __init__(
        self,
        node: Node | None = None,
        pos_sensitivity: float = 1.0,
        rot_sensitivity: float = 3.1415927,
    ):
        if not node:
            rclpy.init()
            self._node = Node("sim", namespace="srb", start_parameter_services=False)  # type: ignore
        else:
            self._node = node

        self.pos_sensitivity = pos_sensitivity
        self.rot_sensitivity = rot_sensitivity

        self.sub_twist = self._node.create_subscription(
            Twist, "/touch/twist_stylus_current", self.cb_twist, 1
        )
        self.sub_button = self._node.create_subscription(
            Bool, "/touch/button", self.cb_button, 1
        )
        self.sub_event = self._node.create_subscription(
            Bool, "/touch/event", self.cb_event, 1
        )
        self.pub_force_feedback = self._node.create_publisher(
            Vector3, "/touch/force_feedback", 1
        )

        self.sub_latency = self._node.create_subscription(
            Float64, "/gui/latency", self.cb_latency, 1
        )
        self.sub_motion_sensitivity = self._node.create_subscription(
            Float64, "/gui/motion_sensitivity", self.cb_motion_sensitivity, 1
        )
        self.sub_force_feedback_sensitivity = self._node.create_subscription(
            Float64,
            "/gui/force_feedback_sensitivity",
            self.cb_force_feedback_sensitivity,
            1,
        )

        # Detects if the button is pressed, which in turn enables the motion
        self.is_motion_enabled = False
        self.gain = 1.0
        self.inactive_feedback_decay = 1.0

        self.latency = 0.0
        self.command_queue = []
        self.feedback_queue = []
        self.last_command = None
        self.last_feedback = None

        self.force_feedback_sensitivity = 1.0

        # Command buffers
        self._close_gripper = False
        self._delta_pos = numpy.zeros(3)  # (x, y, z)
        self._delta_rot = numpy.zeros(3)  # (roll, pitch, yaw)

        # Run a thread for listening to device
        if not node:
            self._executor = MultiThreadedExecutor(num_threads=2)
            self._executor.add_node(self._node)
            self._thread = threading.Thread(target=self._executor.spin)
            self._thread.daemon = True
            self._thread.start()

    def cb_twist(self, msg):
        if self.is_motion_enabled:
            self._delta_pos[0] = -self.pos_sensitivity * self.gain * msg.linear.z
            self._delta_pos[1] = -self.pos_sensitivity * self.gain * msg.linear.x
            self._delta_pos[2] = self.pos_sensitivity * self.gain * msg.linear.y

            self._delta_rot[0] = self.rot_sensitivity * self.gain * msg.angular.z
            self._delta_rot[1] = self.rot_sensitivity * self.gain * msg.angular.x
            self._delta_rot[2] = self.rot_sensitivity * self.gain * msg.angular.y

    def cb_button(self, msg):
        self.is_motion_enabled = msg.data
        if not self.is_motion_enabled:
            self._delta_pos = numpy.zeros(3)
            self._delta_rot = numpy.zeros(3)

    def cb_event(self, msg):
        if msg.data:
            self._close_gripper = not self._close_gripper

    def cb_latency(self, msg):
        if msg.data != self.latency:
            self.latency = msg.data
            self.command_queue = []
            self.feedback_queue = []
            self.last_command = None
            self.last_feedback = None

    def cb_motion_sensitivity(self, msg):
        self.gain = msg.data

    def cb_force_feedback_sensitivity(self, msg):
        self.force_feedback_sensitivity = msg.data

    def __del__(self):
        if hasattr(self, "_thread"):
            self._thread.join()

    def __str__(self) -> str:
        msg = f"Haptic ROS Interface  ({self.__class__.__name__})\n"
        msg += f"Listenining on ROS_DOMAIN_ID: {os.environ.get('ROS_DOMAIN_ID', 0)}\n"
        msg += "Hold to enable motion: Dark grey button\n"
        msg += "Press to toggle the gripper: Light grey button\n"
        return msg

    def reset(self):
        self._close_gripper = False
        self._delta_pos = numpy.zeros(3)
        self._delta_rot = numpy.zeros(3)
        self.command_queue = []
        self.feedback_queue = []
        self.last_command = None
        self.last_feedback = None

    def add_callback(self, key: str, func: Callable):
        raise NotImplementedError

    def advance(self) -> tuple[numpy.ndarray, bool]:
        commands = (
            numpy.concatenate([self._delta_pos, self._delta_rot]),
            self._close_gripper,
        )

        if self.latency == 0.0:
            return commands
        else:
            system_time = self._node.get_clock().now()
            self.command_queue.append((system_time, commands))

            # Find the last viable command
            last_viable_command = None
            for i, (t, _) in enumerate(self.command_queue):
                if (system_time - t).nanoseconds / 1e9 > self.latency:
                    last_viable_command = i
                else:
                    break

            if last_viable_command is not None:
                _, self.last_command = self.command_queue[last_viable_command]
                self.command_queue = self.command_queue[last_viable_command + 1 :]
                return self.last_command
            else:
                if self.last_command is not None:
                    return self.last_command
                else:
                    return numpy.zeros(6), commands[1]

    def set_ft_feedback(self, ft_feedback: numpy.ndarray | torch.Tensor):
        from geometry_msgs.msg import Vector3

        if self.latency == 0.0:
            if self.is_motion_enabled:
                self.pub_force_feedback.publish(
                    Vector3(
                        x=self.force_feedback_sensitivity * float(ft_feedback[0]),
                        y=self.force_feedback_sensitivity * float(ft_feedback[1]),
                        z=self.force_feedback_sensitivity * float(ft_feedback[2]),
                    )
                )
                self.inactive_feedback_decay = 1.0
            else:
                if self.inactive_feedback_decay > 0.0:
                    self.inactive_feedback_decay = max(
                        0.0, 0.1 * self.inactive_feedback_decay
                    )
                    self.pub_force_feedback.publish(
                        Vector3(
                            x=self.inactive_feedback_decay
                            * self.force_feedback_sensitivity
                            * float(ft_feedback[0]),
                            y=self.inactive_feedback_decay
                            * self.force_feedback_sensitivity
                            * float(ft_feedback[1]),
                            z=self.inactive_feedback_decay
                            * self.force_feedback_sensitivity
                            * float(ft_feedback[2]),
                        )
                    )
        else:
            system_time = self._node.get_clock().now()
            self.feedback_queue.append((system_time, ft_feedback))

            # Find the last viable feedback
            last_viable_feedback = None
            for i, (t, _) in enumerate(self.feedback_queue):
                if (system_time - t).nanoseconds / 1e9 > self.latency:
                    last_viable_feedback = i
                else:
                    break

            if last_viable_feedback is not None:
                _, self.last_feedback = self.feedback_queue[last_viable_feedback]
                self.feedback_queue = self.feedback_queue[last_viable_feedback + 1 :]
                self.pub_force_feedback.publish(
                    Vector3(
                        x=self.force_feedback_sensitivity
                        * float(self.last_feedback[0]),
                        y=self.force_feedback_sensitivity
                        * float(self.last_feedback[1]),
                        z=self.force_feedback_sensitivity
                        * float(self.last_feedback[2]),
                    )
                )



import math
from typing import Sequence, Tuple

import torch
from isaaclab.utils.math import *  # noqa: F403  # type: ignore
from isaaclab.utils.math import combine_frame_transforms, matrix_from_quat
from isaaclab.utils.math import subtract_frame_transforms as _subtract_frame_transforms
from isaaclab.utils.math import transform_points as _transform_points


@torch.jit.script
def deg_to_rad(deg: float) -> float:
    return deg * math.pi / 180.0


@torch.jit.script
def rad_to_deg(rad: float) -> float:
    return rad * 180.0 / math.pi


def rpy_to_quat(
    *rpy: float | Sequence[float], deg: bool = True
) -> Tuple[float, float, float, float]:
    """
    Returns wxyz quaternion from roll-pitch-yaw angles.
    Accepts either separate values or a sequence of 3 numbers.

    Args:
        rpy: Either (roll, pitch, yaw) as separate floats or a sequence of 3 numbers
        deg: If True, input is in degrees, otherwise in radians

    Returns:
        Tuple of (w, x, y, z) quaternion components

    Raises:
        ValueError: If input doesn't contain exactly 3 values
    """
    if len(rpy) == 3:
        roll, pitch, yaw = rpy
    elif len(rpy) == 1 and isinstance(rpy[0], Sequence) and len(rpy[0]) == 3:
        roll, pitch, yaw = rpy[0]
    else:
        raise ValueError(
            "Input must be either 3 separate values or a sequence of 3 values"
        )

    if roll == 0.0 and pitch == 0.0 and yaw == 0.0:
        return 1.0, 0.0, 0.0, 0.0

    if deg:
        roll, pitch, yaw = (deg_to_rad(angle) for angle in (roll, pitch, yaw))  # type: ignore

    # Compute half angles
    r2, p2, y2 = roll / 2.0, pitch / 2.0, yaw / 2.0  # type: ignore

    # Precompute trig functions
    cr, cp, cy = math.cos(r2), math.cos(p2), math.cos(y2)
    sr, sp, sy = math.sin(r2), math.sin(p2), math.sin(y2)

    return (
        cy * cr * cp + sy * sr * sp,  # w
        cy * sr * cp - sy * cr * sp,  # x
        cy * cr * sp + sy * sr * cp,  # y
        sy * cr * cp - cy * sr * sp,  # z
    )


@torch.jit.script
def rotmat_to_rot6d(rotmat: torch.Tensor) -> torch.Tensor:
    return rotmat[..., :, :2].reshape(rotmat.shape[:-2] + (6,))


@torch.jit.script
def quat_to_rot6d(quaternions: torch.Tensor) -> torch.Tensor:
    return rotmat_to_rot6d(matrix_from_quat(quaternions))


def combine_frame_transforms_tuple(
    t01: Tuple[float, float, float] | Sequence[float],
    q01: Tuple[float, float, float, float] | Sequence[float],
    t12: Tuple[float, float, float] | Sequence[float] | None = None,
    q12: Tuple[float, float, float, float] | Sequence[float] | None = None,
) -> Tuple[Tuple[float, float, float], Tuple[float, float, float, float]]:
    _t01 = torch.tensor(t01)
    _q01 = torch.tensor(q01)
    _t12 = torch.tensor(t12) if t12 is not None else None
    _q12 = torch.tensor(q12) if q12 is not None else None
    t02, q02 = combine_frame_transforms(_t01, _q01, _t12, _q12)
    return tuple(t02.cpu().tolist()), tuple(q02.cpu().tolist())


@torch.jit.script
def subtract_frame_transforms(
    t01: torch.Tensor,
    q01: torch.Tensor,
    t02: torch.Tensor | None = None,
    q02: torch.Tensor | None = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    return _subtract_frame_transforms(t01, q01, t02, q02)


def subtract_frame_transforms_tuple(
    t01: Tuple[float, float, float],
    q01: Tuple[float, float, float, float],
    t02: Tuple[float, float, float] | None = None,
    q02: Tuple[float, float, float, float] | None = None,
) -> Tuple[Tuple[float, float, float], Tuple[float, float, float, float]]:
    _t01 = torch.tensor(t01)
    _q01 = torch.tensor(q01)
    _t02 = torch.tensor(t02) if t02 is not None else None
    _q02 = torch.tensor(q02) if q02 is not None else None
    t12, q12 = subtract_frame_transforms(_t01, _q01, _t02, _q02)
    return tuple(t12.cpu().tolist()), tuple(q12.cpu().tolist())


@torch.jit.script
def transform_points(
    points: torch.Tensor,
    pos: torch.Tensor | None = None,
    quat: torch.Tensor | None = None,
) -> torch.Tensor:
    return _transform_points(points, pos, quat)



from srb.core.asset import AssetBaseCfg, Terrain
from srb.core.sim import GroundPlaneCfg


class GroundPlane(Terrain):
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="/World/ground_plane",
        spawn=GroundPlaneCfg(
            color=(0.0, 158.0 / 255.0, 218.0 / 255.0),
        ),
    )



from srb.core.action.action_group import ActionGroup
from srb.core.action.term import DummyActionCfg
from srb.utils.cfg import configclass


@configclass
class DummyActionGroup(ActionGroup):
    dummy: DummyActionCfg = DummyActionCfg()



from torch import pi

from srb.core.action import (
    ActionGroup,
    MulticopterBodyAccelerationActionCfg,
    MulticopterBodyAccelerationActionGroup,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, Multicopter, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class Ingenuity(Multicopter):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/ingenuity",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("copter")
            .joinpath("ingenuity.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=1,
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
        ),
        actuators={
            "rotors": ImplicitActuatorCfg(
                joint_names_expr=["rotor_joint_[1-2]"],
                velocity_limit=2500 / 60 * 2 * pi,
                effort_limit=7.5,
                stiffness=0.0,
                damping=1000.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = MulticopterBodyAccelerationActionGroup(
        MulticopterBodyAccelerationActionCfg(
            asset_name="robot",
            frame_base="body",
            regex_rotor_joints="rotor_joint_[1-2]",
            nominal_rpm={
                "rotor_joint_1": 2500.0,
                "rotor_joint_2": -2500.0,
            },
            tilt_magnitude=0.125,
            scale=0.5,
        )
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="body")
    frame_payload_mount: Frame = Frame(
        prim_relpath="body",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="body",
        offset=Transform(
            pos=(0.0, 0.0, 0.13),
            rot=rpy_to_quat(0.0, 180.0, 0.0),
        ),
    )
    frame_onboard_camera: Frame = Frame(
        prim_relpath="body/camera_onboard",
        offset=Transform(
            pos=(0.045, 0.0, 0.1275),
            rot=rpy_to_quat(0.0, 90.0, 0.0),
        ),
    )



from typing import TYPE_CHECKING

import isaacsim.core.utils.prims as prim_utils
from isaaclab.sim import bind_physics_material, bind_visual_material, clone
from pxr import Usd

from srb.core.sim import schemas

if TYPE_CHECKING:
    from . import cfg


@clone
def spawn_arrow(
    prim_path: str,
    cfg: "cfg.ArrowCfg",
    translation: tuple[float, float, float] | None = None,
    orientation: tuple[float, float, float, float] | None = None,
) -> Usd.Prim:
    if prim_utils.is_prim_path_valid(prim_path):
        raise ValueError(f"A prim already exists at path: '{prim_path}'.")

    prim = prim_utils.create_prim(
        prim_path, "Xform", translation=translation, orientation=orientation
    )

    geom_prim_path = prim_path + "/geometry"
    mesh_prim_path = geom_prim_path + "/mesh"

    container = prim_utils.create_prim(mesh_prim_path, "Xform")

    arrow_body_attributes = {
        "radius": cfg.tail_radius,
        "height": cfg.tail_length,
        "axis": "X",
    }
    arrow_body_translation = (cfg.tail_length / 2.0, 0, 0)
    arrow_body = prim_utils.create_prim(
        str(container.GetPath().AppendChild("arrow_body")),
        "Cylinder",
        position=arrow_body_translation,
        attributes=arrow_body_attributes,
    )

    arrow_head_attributes = {
        "radius": cfg.head_radius,
        "height": cfg.head_length,
        "axis": "X",
    }
    arrow_head_translation = (cfg.tail_length + cfg.head_length / 2.0, 0, 0)
    arrow_head = prim_utils.create_prim(
        str(container.GetPath().AppendChild("arrow_head")),
        "Cone",
        position=arrow_head_translation,
        attributes=arrow_head_attributes,
    )

    if cfg.collision_props is not None:
        schemas.define_collision_properties(
            str(arrow_body.GetPath()), cfg.collision_props
        )
        schemas.define_collision_properties(
            str(arrow_head.GetPath()), cfg.collision_props
        )

    if cfg.visual_material is not None:
        if not cfg.visual_material_path.startswith("/"):
            material_path = f"{geom_prim_path}/{cfg.visual_material_path}"
        else:
            material_path = cfg.visual_material_path
        cfg.visual_material.func(material_path, cfg.visual_material)
        bind_visual_material(mesh_prim_path, material_path)

    if cfg.physics_material is not None:
        if not cfg.physics_material_path.startswith("/"):
            material_path = f"{geom_prim_path}/{cfg.physics_material_path}"
        else:
            material_path = cfg.physics_material_path
        cfg.physics_material.func(material_path, cfg.physics_material)
        bind_physics_material(mesh_prim_path, material_path)

    return prim


@clone
def spawn_pinned_arrow(
    prim_path: str,
    cfg: "cfg.PinnedArrowCfg",
    translation: tuple[float, float, float] | None = None,
    orientation: tuple[float, float, float, float] | None = None,
) -> Usd.Prim:
    if prim_utils.is_prim_path_valid(prim_path):
        raise ValueError(f"A prim already exists at path: '{prim_path}'.")

    prim = prim_utils.create_prim(
        prim_path, "Xform", translation=translation, orientation=orientation
    )

    geom_prim_path = prim_path + "/geometry"
    mesh_prim_path = geom_prim_path + "/mesh"

    container = prim_utils.create_prim(mesh_prim_path, "Xform")

    pin_attributes = {"radius": cfg.pin_radius, "height": cfg.pin_length, "axis": "Z"}
    pin_translation = (0, 0, cfg.pin_length / 2.0)
    pin = prim_utils.create_prim(
        str(container.GetPath().AppendChild("pin_body")),
        "Cylinder",
        position=pin_translation,
        attributes=pin_attributes,
    )

    arrow_body_attributes = {
        "radius": cfg.tail_radius,
        "height": cfg.tail_length,
        "axis": "X",
    }
    arrow_body_translation = (cfg.tail_length / 2.0, 0, cfg.pin_length)
    arrow_body = prim_utils.create_prim(
        str(container.GetPath().AppendChild("arrow_body")),
        "Cylinder",
        position=arrow_body_translation,
        attributes=arrow_body_attributes,
    )

    arrow_head_attributes = {
        "radius": cfg.head_radius,
        "height": cfg.head_length,
        "axis": "X",
    }
    arrow_head_translation = (
        cfg.tail_length + cfg.head_length / 2.0,
        0,
        cfg.pin_length,
    )
    arrow_head = prim_utils.create_prim(
        str(container.GetPath().AppendChild("arrow_head")),
        "Cone",
        position=arrow_head_translation,
        attributes=arrow_head_attributes,
    )

    if cfg.collision_props is not None:
        schemas.define_collision_properties(
            str(arrow_body.GetPath()), cfg.collision_props
        )
        schemas.define_collision_properties(
            str(arrow_head.GetPath()), cfg.collision_props
        )
        schemas.define_collision_properties(str(pin.GetPath()), cfg.collision_props)

    if cfg.visual_material is not None:
        if not cfg.visual_material_path.startswith("/"):
            material_path = f"{geom_prim_path}/{cfg.visual_material_path}"
        else:
            material_path = cfg.visual_material_path
        cfg.visual_material.func(material_path, cfg.visual_material)
        bind_visual_material(mesh_prim_path, material_path)

    if cfg.physics_material is not None:
        if not cfg.physics_material_path.startswith("/"):
            material_path = f"{geom_prim_path}/{cfg.physics_material_path}"
        else:
            material_path = cfg.physics_material_path
        cfg.physics_material.func(material_path, cfg.physics_material)
        bind_physics_material(mesh_prim_path, material_path)

    return prim


@clone
def spawn_pinned_sphere(
    prim_path: str,
    cfg: "cfg.PinnedSphereCfg",
    translation: tuple[float, float, float] | None = None,
    orientation: tuple[float, float, float, float] | None = None,
) -> Usd.Prim:
    if prim_utils.is_prim_path_valid(prim_path):
        raise ValueError(f"A prim already exists at path: '{prim_path}'.")

    prim = prim_utils.create_prim(
        prim_path, "Xform", translation=translation, orientation=orientation
    )

    geom_prim_path = prim_path + "/geometry"
    mesh_prim_path = geom_prim_path + "/mesh"

    container = prim_utils.create_prim(mesh_prim_path, "Xform")

    pin_attributes = {"radius": cfg.pin_radius, "height": cfg.pin_length, "axis": "Z"}
    pin_translation = (0, 0, cfg.pin_length / 2.0)
    pin = prim_utils.create_prim(
        str(container.GetPath().AppendChild("pin_body")),
        "Cylinder",
        position=pin_translation,
        attributes=pin_attributes,
    )
    sphere_attributes = {"radius": cfg.sphere_radius}
    sphere_translation = (0, 0, cfg.pin_length)
    sphere = prim_utils.create_prim(
        str(container.GetPath().AppendChild("pin_head")),
        "Sphere",
        position=sphere_translation,
        attributes=sphere_attributes,
    )

    if cfg.collision_props is not None:
        schemas.define_collision_properties(str(sphere.GetPath()), cfg.collision_props)
        schemas.define_collision_properties(str(pin.GetPath()), cfg.collision_props)

    if cfg.visual_material is not None:
        if not cfg.visual_material_path.startswith("/"):
            material_path = f"{geom_prim_path}/{cfg.visual_material_path}"
        else:
            material_path = cfg.visual_material_path
        cfg.visual_material.func(material_path, cfg.visual_material)
        bind_visual_material(mesh_prim_path, material_path)

    if cfg.physics_material is not None:
        if not cfg.physics_material_path.startswith("/"):
            material_path = f"{geom_prim_path}/{cfg.physics_material_path}"
        else:
            material_path = cfg.physics_material_path
        cfg.physics_material.func(material_path, cfg.physics_material)
        bind_physics_material(mesh_prim_path, material_path)

    return prim



try:
    import rclpy
except ImportError:
    from srb.utils.ros import enable_ros2_bridge

    enable_ros2_bridge()

import threading
from queue import Queue
from typing import TYPE_CHECKING

import numpy
import rclpy
from pxr import Gf
from rclpy.executors import MultiThreadedExecutor
from rclpy.node import Node
from std_msgs.msg import Bool, Empty, Float64

from .base import InterfaceBase

if TYPE_CHECKING:
    from srb._typing import AnyEnv


class GuiInterface(InterfaceBase):
    def __init__(self, env: "AnyEnv", node: Node | None = None, *args, **kwargs):
        self._env = env

        ## Initialize node
        if not node:
            rclpy.init()
            self._node = Node("sim", namespace="srb", start_parameter_services=False)  # type: ignore
        else:
            self._node = node

        ## Execution queue for actions and services that must be executed in the main thread between environment steps via `update()`
        self._exec_queue = Queue()

        ## Subscribers
        self._sub_reset = self._node.create_subscription(
            Empty, "/gui/reset_discard_dataset", self._cb_reset, 1
        )
        self._sub_shutdown_process = self._node.create_subscription(
            Empty, "/gui/shutdown_process", self._cb_shutdown_process, 1
        )
        self._sub_gravity = self._node.create_subscription(
            Float64, "/gui/gravity", self._cb_gravity, 1
        )

        # Run a thread for listening to device
        if not node:
            self._executor = MultiThreadedExecutor(num_threads=2)
            self._executor.add_node(self._node)
            self._thread = threading.Thread(target=self._executor.spin)
            self._thread.daemon = True
            self._thread.start()

    def __del__(self):
        if hasattr(self, "_thread"):
            self._thread.join()

    def update(self, *args, **kwargs):
        while not self._exec_queue.empty():
            request, request_kwargs = self._exec_queue.get()
            request(**request_kwargs)

    def reset(self):
        self._env.reset()

    def shutdown(self):
        exit(0)

    def set_gravity(self, gravity: float):
        self._env.unwrapped.sim.cfg.gravity = (0.0, 0.0, -gravity)  # type: ignore

        physics_scene = self._env.unwrapped.sim._physics_context._physics_scene  # type: ignore

        gravity = numpy.asarray(self._env.unwrapped.sim.cfg.gravity)  # type: ignore
        gravity_magnitude = numpy.linalg.norm(gravity)

        # Avoid division by zero
        if gravity_magnitude != 0.0:
            gravity_direction = gravity / gravity_magnitude
        else:
            gravity_direction = gravity

        physics_scene.CreateGravityDirectionAttr(Gf.Vec3f(*gravity_direction))
        physics_scene.CreateGravityMagnitudeAttr(gravity_magnitude)

    def _cb_reset(self, msg: Bool):
        self._exec_queue.put((self.reset, {}))

    def _cb_shutdown_process(self, msg: Bool):
        self._exec_queue.put((self.shutdown, {}))

    def _cb_gravity(self, msg: Float64):
        self._exec_queue.put((self.set_gravity, {"gravity": msg.data}))



import logging
import sys
from os import environ

from srb.utils.tracing import with_logfire, with_rich

# Define TRACE log level
TRACE: int = 5
logging.addLevelName(TRACE, "TRACE")

# Create a logger
logger = logging.getLogger("srb")
logger.propagate = False

# Set the logger level
logger.setLevel(
    (environ.get("LOG_LEVEL") or environ.get("SRB_LOG_LEVEL", "INFO")).upper()
)

# Set up console logging handlers (either rich or plain)
if with_rich():
    from rich.logging import RichHandler

    logger.addHandler(
        RichHandler(
            omit_repeated_times=False,
            rich_tracebacks=True,
            log_time_format="[%X]",
        )
    )
else:
    # Create a handler for logs below WARNING (DEBUG and INFO) to STDOUT
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)  # Allow DEBUG and INFO
    stdout_handler.addFilter(lambda record: record.levelno < logging.WARNING)

    # Create a handler for logs WARNING and above (WARNING, ERROR, CRITICAL) to STDERR
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(logging.WARNING)  # Allow WARNING and above

    # Define a formatter and set it for both handlers
    formatter = logging.Formatter(
        fmt="[{asctime}] {levelname:8s} {message}",
        datefmt="%H:%M:%S",
        style="{",
    )
    stdout_handler.setFormatter(formatter)
    stderr_handler.setFormatter(formatter)

    # Add handlers to the logger
    logger.addHandler(stdout_handler)
    logger.addHandler(stderr_handler)

# Set up Logfire logging handler
if with_logfire():
    from logfire import LogfireLoggingHandler

    logger.addHandler(LogfireLoggingHandler())


__all__ = [
    "logger",
    "trace",
    "debug",
    "info",
    "warning",
    "error",
    "critical",
]

debug = logger.debug
info = logger.info
warning = logger.warning
error = logger.error
critical = logger.critical


def trace(msg, *args, **kwargs):
    if logger.isEnabledFor(TRACE):  # type: ignore
        logger._log(TRACE, msg, args, **kwargs)  # type: ignore



from srb.core.marker import VisualizationMarkersCfg
from srb.core.sim import ArrowCfg

ARROW_CFG = VisualizationMarkersCfg(
    markers={
        "arrow": ArrowCfg(
            tail_length=0.2,
            tail_radius=0.05,
            head_radius=0.1,
            head_length=0.15,
        ),
    }
)



from srb.core.action import (
    ActionGroup,
    JointPositionActionCfg,
    JointPositionActionGroup,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, Humanoid, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import rpy_to_quat
from srb.utils.nucleus import ISAAC_NUCLEUS_DIR, ISAACLAB_NUCLEUS_DIR


class Humanoid21(Humanoid):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/humanoid",
        spawn=UsdFileCfg(
            usd_path=f"{ISAAC_NUCLEUS_DIR}/Robots/Humanoid/humanoid_instanceable.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=None,
                max_depenetration_velocity=10.0,
                enable_gyroscopic_forces=True,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 1.34),
            joint_pos={".*": 0.0},
        ),
        actuators={
            "body": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                stiffness={
                    ".*_waist.*": 20.0,
                    ".*_upper_arm.*": 10.0,
                    "pelvis": 10.0,
                    ".*_lower_arm": 2.0,
                    ".*_thigh:0": 10.0,
                    ".*_thigh:1": 20.0,
                    ".*_thigh:2": 10.0,
                    ".*_shin": 5.0,
                    ".*_foot.*": 2.0,
                },
                damping={
                    ".*_waist.*": 5.0,
                    ".*_upper_arm.*": 5.0,
                    "pelvis": 5.0,
                    ".*_lower_arm": 1.0,
                    ".*_thigh:0": 5.0,
                    ".*_thigh:1": 5.0,
                    ".*_thigh:2": 5.0,
                    ".*_shin": 0.1,
                    ".*_foot.*": 1.0,
                },
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="pelvis")
    frame_front_camera: Frame = Frame(
        prim_relpath="pelvis/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = "(left|right)_foot"


class Humanoid28(Humanoid):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/humanoid_28",
        spawn=UsdFileCfg(
            usd_path=f"{ISAACLAB_NUCLEUS_DIR}/Robots/Classic/Humanoid28/humanoid_28.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=None,
                max_depenetration_velocity=10.0,
                enable_gyroscopic_forces=True,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.8),
            joint_pos={".*": 0.0},
        ),
        actuators={
            "body": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                stiffness=None,
                damping=None,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="pelvis")
    frame_front_camera: Frame = Frame(
        prim_relpath="pelvis/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = "(left|right)_foot"



from typing import Tuple

from pydantic import BaseModel


class Transform(BaseModel):
    pos: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    rot: Tuple[float, float, float, float] = (1.0, 0.0, 0.0, 0.0)



from srb.assets.object.tool import FrankaHand
from srb.core.action import (  # noqa: F401
    ActionGroup,
    DifferentialIKControllerCfg,
    DifferentialInverseKinematicsActionCfg,
    InverseKinematicsActionGroup,
    JointEffortActionCfg,
    JointEffortActionGroup,
    JointPositionRelativeActionGroup,
    OperationalSpaceControlActionGroup,
    OperationalSpaceControllerActionCfg,
    OperationalSpaceControllerCfg,
    RelativeJointPositionActionCfg,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, SerialManipulator, Tool, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


### ANCHOR: example_p1 (docs)
class Franka(SerialManipulator):
    ## Model - Articulation with several links connected by joints
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/franka",
        ## Spawner loads a static USD file
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("manipulator")
            .joinpath("franka_arm.usdz")
            .as_posix(),
            ### ANCHOR_END: example_p1 (docs)
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
            ### ANCHOR: example_p2 (docs)
        ),
        ## Initial joint configuration of the robot
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "panda_joint1": 0.0,
                "panda_joint2": 0.0,
                "panda_joint3": 0.0,
                "panda_joint4": deg_to_rad(-90.0),
                "panda_joint5": 0.0,
                "panda_joint6": deg_to_rad(90.0),
                "panda_joint7": deg_to_rad(45.0),
            },
        ),
        ### ANCHOR_END: example_p2 (docs)
        actuators={
            "panda_shoulder": ImplicitActuatorCfg(
                joint_names_expr=["panda_joint[1-4]"],
                effort_limit=87.0,
                velocity_limit=2.175,
                stiffness=4000.0,
                damping=800.0,
            ),
            "panda_forearm": ImplicitActuatorCfg(
                joint_names_expr=["panda_joint[5-7]"],
                effort_limit=12.0,
                velocity_limit=2.61,
                stiffness=4000.0,
                damping=800.0,
            ),
        },
        ### ANCHOR: example_p3 (docs)
    )
    ## End effector - The default hand is separate to allow for easy replacement
    end_effector: Tool | None = FrankaHand()

    ## Actions - Inverse Kinematics action group that drives all joints
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=["panda_joint[1-7]"],
            base_name="panda_link0",
            body_name="panda_link7",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.1,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    ### ANCHOR_END: example_p3 (docs)
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=["panda_joint[1-7]"],
    #         body_name="panda_link7",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="fixed",
    #             motion_stiffness_task=100.0,
    #             motion_damping_ratio_task=1.0,
    #             # motion_stiffness_task=250.0,
    #             # motion_damping_ratio_task=1.5,
    #             nullspace_control="position",
    #         ),
    #         nullspace_joint_pos_target="center",
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=["panda_joint[1-7]"],
    #         body_name="panda_link7",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #             # motion_damping_ratio_task=1.5,
    #             nullspace_control="position",
    #         ),
    #         nullspace_joint_pos_target="center",
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=["panda_joint[1-7]"],
    #         body_name="panda_link7",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable",
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_limits_task=(0.5, 2.5),
    #             nullspace_control="position",
    #         ),
    #         nullspace_joint_pos_target="center",
    #         position_scale=0.1,
    #         orientation_scale=0.1,
    #         stiffness_scale=120.0,
    #         damping_ratio_scale=1.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ### ANCHOR: example_p4 (docs)
    ## Frames - Relevant frames for attaching the robot and mounting tool/sensors
    frame_base: Frame = Frame(prim_relpath="panda_link0")
    frame_flange: Frame = Frame(
        prim_relpath="panda_link7",
        offset=Transform(
            pos=(0.0, 0.0, 0.107),
            rot=rpy_to_quat(0.0, 0.0, -45.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="panda_link7/camera_wrist",
        offset=Transform(
            pos=(0.075, -0.075, 0.1),
            rot=rpy_to_quat(0.0, -80.0, 135.0),
        ),
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="panda_link0/camera_base",
        offset=Transform(
            pos=(0.15, 0.0, 0.1),
            rot=rpy_to_quat(0.0, 45.0, 0.0),
        ),
    )
    ### ANCHOR_END: example_p4 (docs)



from typing import Dict, NamedTuple, Tuple, TypeAlias

from torch import Tensor

_StepReturn: TypeAlias = Tuple[
    Dict[
        str,
        Dict[str, Tensor],
    ],
    Dict[str, Tensor],
    Tensor,
    Tensor,
]


class StepReturn(NamedTuple):
    observation: Dict[str, Dict[str, Tensor]]
    reward: Dict[str, Tensor]
    termination: Tensor
    truncation: Tensor
    info: Dict[str, Tensor] | None = None

    @staticmethod
    def from_tuple(step_return: _StepReturn) -> "StepReturn":
        return StepReturn(*step_return)



from typing import TYPE_CHECKING, Tuple

from isaaclab.sim.spawners.wrappers import MultiAssetSpawnerCfg, spawn_multi_asset
from pxr import Usd

from srb.core.sim import CapsuleCfg, ConeCfg, CuboidCfg, CylinderCfg, SphereCfg

if TYPE_CHECKING:
    from .cfg import MultiShapeSpawnerCfg

IGNORED_SPAWN_ATTRIBUTES = (
    "func",
    "shapes",
    "scale",
    "radius",
    "height",
    "axis",
    "random_choice",
)


def spawn_multi_shape(
    prim_path: str,
    cfg: "MultiShapeSpawnerCfg",
    translation: Tuple[float, float, float] | None = None,
    orientation: Tuple[float, float, float, float] | None = None,
) -> Usd.Prim:
    # Collect shape spawner kwargs
    shape_cfg_kwargs = {
        attr_name: attr_value
        for attr_name, attr_value in cfg.__dict__.items()
        if attr_name not in IGNORED_SPAWN_ATTRIBUTES
    }

    # Collect shape spawner configs
    assets_cfg = []
    if not cfg.shapes or "cuboid" in cfg.shapes:
        assets_cfg.append(CuboidCfg(size=cfg.scale, **shape_cfg_kwargs))
    if not cfg.shapes or "sphere" in cfg.shapes:
        assets_cfg.append(
            SphereCfg(radius=cfg.radius or cfg.scale[0], **shape_cfg_kwargs)
        )
    if not cfg.shapes or "cylinder" in cfg.shapes:
        assets_cfg.append(
            CylinderCfg(
                radius=cfg.radius or cfg.scale[0],
                height=cfg.height or cfg.scale[1],
                axis=cfg.axis,
                **shape_cfg_kwargs,
            )
        )
    if not cfg.shapes or "capsule" in cfg.shapes:
        assets_cfg.append(
            CapsuleCfg(
                radius=cfg.radius or cfg.scale[0],
                height=cfg.height or cfg.scale[1],
                axis=cfg.axis,
                **shape_cfg_kwargs,
            )
        )
    if not cfg.shapes or "cone" in cfg.shapes:
        assets_cfg.append(
            ConeCfg(
                radius=cfg.radius or cfg.scale[0],
                height=cfg.height or cfg.scale[1],
                axis=cfg.axis,
                **shape_cfg_kwargs,
            )
        )

    # Create and spawn multi-asset configuration
    return spawn_multi_asset(
        prim_path=prim_path,
        cfg=MultiAssetSpawnerCfg(
            assets_cfg=assets_cfg,
            random_choice=cfg.random_choice,
        ),
        translation=translation,
        orientation=orientation,
    )



from dataclasses import MISSING

from srb import assets
from srb.core.action import (
    InverseKinematicsActionGroup,
    JointPositionRelativeActionGroup,
    OperationalSpaceControlActionGroup,
)
from srb.core.asset import Articulation, AssetVariant, GroundManipulator
from srb.core.env.manipulation.env import (
    ManipulationEnv,
    ManipulationEventCfg,
    ManipulationSceneCfg,
)
from srb.core.env.mobile.ground.env import (
    GroundEnv,
    GroundEnvCfg,
    GroundEventCfg,
    GroundSceneCfg,
)
from srb.core.manager import SceneEntityCfg
from srb.core.sensor import ContactSensorCfg
from srb.utils.cfg import configclass
from srb.utils.math import combine_frame_transforms_tuple


@configclass
class GroundManipulationSceneCfg(GroundSceneCfg, ManipulationSceneCfg):
    pass


@configclass
class GroundManipulationEventCfg(GroundEventCfg, ManipulationEventCfg):
    pass


@configclass
class GroundManipulationEnvCfg(GroundEnvCfg):
    ## Assets
    robot: GroundManipulator | AssetVariant = assets.GenericGroundManipulator(
        mobile_base=assets.Spot(), manipulator=assets.Franka()
    )
    _robot: GroundManipulator = MISSING  # type: ignore

    ## Scene
    scene: GroundManipulationSceneCfg = GroundManipulationSceneCfg()

    ## Events
    events: GroundManipulationEventCfg = GroundManipulationEventCfg()

    ## Time
    env_rate: float = 1.0 / 150.0
    agent_rate: float = 1.0 / 50.0

    def __post_init__(self):
        ## Jacobian-based actions are currently not supported for free-floating manipulators
        if isinstance(self.robot, GroundManipulator) and isinstance(
            self.robot.manipulator.actions,
            (InverseKinematicsActionGroup, OperationalSpaceControlActionGroup),
        ):
            self.robot.manipulator.actions = JointPositionRelativeActionGroup()

        super().__post_init__()
        assert self.scene.manipulator is not None

        ## Adapted from ManipulationEnvCfg
        # Sensor: End-effector transform
        self.scene.tf_end_effector.prim_path = f"{self.scene.manipulator.prim_path}/{self._robot.manipulator.frame_base.prim_relpath}"
        self.scene.tf_end_effector.target_frames[
            0
        ].prim_path = f"{self.scene.manipulator.prim_path}/{self._robot.manipulator.frame_flange.prim_relpath}"
        if self._robot.manipulator.end_effector is not None:
            (
                self.scene.tf_end_effector.target_frames[0].offset.pos,
                self.scene.tf_end_effector.target_frames[0].offset.rot,
            ) = combine_frame_transforms_tuple(
                self._robot.manipulator.frame_flange.offset.pos,
                self._robot.manipulator.frame_flange.offset.rot,
                self._robot.manipulator.end_effector.frame_tool_centre_point.offset.pos,
                self._robot.manipulator.end_effector.frame_tool_centre_point.offset.rot,
            )
        else:
            (
                self.scene.tf_end_effector.target_frames[0].offset.pos,
                self.scene.tf_end_effector.target_frames[0].offset.rot,
            ) = (
                self._robot.manipulator.frame_flange.offset.pos,
                self._robot.manipulator.frame_flange.offset.rot,
            )

        # Sensor: Robot contacts
        self.scene.contacts_robot.prim_path = f"{self.scene.manipulator.prim_path}/.*"

        # Sensor: End-effector contacts
        self.scene.contacts_end_effector = (
            ContactSensorCfg(
                prim_path=f"{self._robot.manipulator.end_effector.asset_cfg.prim_path}/.*",
            )
            if self._robot.manipulator.end_effector is not None
            else None
        )

        # Event: Randomize robot joints
        self.events.randomize_robot_joints.params["asset_cfg"] = SceneEntityCfg(
            "manipulator"
        )


class GroundManipulationEnv(GroundEnv, ManipulationEnv):
    cfg: GroundManipulationEnvCfg

    def __init__(self, cfg: GroundManipulationEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._manipulator: Articulation = self.scene["manipulator"]



from typing import TYPE_CHECKING, Type

import torch

from srb.core.manager import ActionTerm, ActionTermCfg
from srb.utils.cfg import configclass

if TYPE_CHECKING:
    from srb._typing import AnyEnv
    from srb.core.asset import Articulation, RigidObject


class DummyAction(ActionTerm):
    cfg: "DummyActionCfg"
    _env: "AnyEnv"
    _asset: "Articulation | RigidObject"

    @property
    def action_dim(self) -> int:
        return self.cfg.dim

    @property
    def raw_actions(self) -> torch.Tensor:
        return self._raw_actions

    @property
    def processed_actions(self) -> torch.Tensor:
        return self._processed_actions

    def process_actions(self, actions):
        self._raw_actions = actions
        self._processed_actions = self._raw_actions

    def apply_actions(self):
        pass


@configclass
class DummyActionCfg(ActionTermCfg):
    class_type: Type = DummyAction
    asset_name: str = "robot"
    dim: int = 1



from srb.core.asset import Object, RigidObjectCfg
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MultiShapeSpawnerCfg,
    RigidBodyPropertiesCfg,
)


class RandomShape(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/shape",
        spawn=MultiShapeSpawnerCfg(
            collision_props=CollisionPropertiesCfg(),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=1000.0),
        ),
    )



from collections.abc import Callable
from typing import Literal, Sequence, Tuple

from srb.core.sim import ShapeCfg
from srb.utils.cfg import configclass

from .impl import spawn_multi_shape


@configclass
class MultiShapeSpawnerCfg(ShapeCfg):
    func: Callable = spawn_multi_shape

    shapes: Sequence[Literal["cuboid", "sphere", "cylinder", "capsule", "cone"]] = ()
    """Shapes to spawn (keep empty to consider all shapes)"""

    scale: Tuple[float, float, float] = (1.0, 1.0, 1.0)
    """Scale of cuboid (affects other shapes if radius and height remain unset)"""

    radius: float | None = None
    """Radius of sphere|cylinder|capsule|cone (default: self.scale[0])"""

    height: float | None = None
    """Height of cylinder|capsule|cone (default: self.scale[1])"""

    axis: Literal["X", "Y", "Z"] = "Z"
    """Axis of cylinder|capsule|cone"""

    random_choice: bool = True
